/*
** Copyright (c) 2005-2008, 2010-2011 7B Software, Inc.
** All rights reserved.
**
** This file is part of XMLP.
**
** XMLP is free software: you can redistribute it and/or modify
** it under the terms of the GNU Lesser General Public License as published by
** the Free Software Foundation, either version 3 of the License, or
** (at your option) any later version.
**
** XMLP is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Lesser General Public License for more details.
**
** You should have received a copy of the GNU Lesser General Public License
** along with XMLP.  If not, see <http://www.gnu.org/licenses/>.
**
** Author contact information:
**   mark.smith@7bsoftware.com
**   http://www.7bsoftware.com
*/

/*
** A library to parse and manage an XML database.
**
** This is not DOM.  DOM is way too complicated for what
** we need.
**
** This library actually processes a subset of XML as follows:
**
**    *  The content of an element must be either other elements or
**       CDATA, never both.
**    *  The order of elements is arbitrary.
**    *  There are no processing instructions.
**
** With these restrictions, we can model the XML like a hierarchical
** filesystem.
*/
static const char rcsid[] = "@(#) $Header: /home/cvs/softools/code/common/xmlp/src/xmlp.cpp,v 1.69 2011/05/21 01:48:16 eas Exp $";
#include <stdio.h>
#include <assert.h>
#include <string.h>
#include <stdlib.h>
#include <ctype.h>
#include <math.h>
#include <errno.h>
#include <limits.h>
#include <tcl.h>
#include "xmlp.h"

#if INTERFACE

/*
** object: XmlNode
**
** The XML is represented as a tree of XmlNode objects.  Client code
** may not access XmlNode objects directly.  Only an XmlCursor object
** is permitted to access an XmlNode object.  Client code uses the
** XmlCursor object interface and the XmlCursor manipulates XmlNodes
** as necessary.
**
** Each XmlNode represents a single XML element.
** Each XmlNode contains pointers to its parent node, pointers to its
** left and right sibling nodes (all sibling elements are stored in
** a balanced binary tree) and a pointer to the root of a binary
** tree of child nodes.  An XmlNode can contain data (stored as either
** text or as a floating point number) or a tree of child XmlNodes
** but not both.
*/
class XmlNode {
private:
  XmlNode *pParent;        // The parent node.  NULL if this is the root
  XmlNode *pUp;            // Sibling above us in the b-tree of siblings
  XmlNode *pLeft;          // Sibling to the left.  Less than "this".
  XmlNode *pRight;         // Sibling to the right.  Greater than "this".
  XmlNode *pChild;         // Top of the b-tree of children
  int weight;              // pLeft->weight + pRight->weight + 1
  char *zName;             // Name (aka: element tag) for this node
  char *zFull;             // Full pathname of self
  char *zBase;             // zName with :index appended
  int iIndex;              // Index to make zName unique
  char *zData;             // CDATA content for this node.
  int nData;               // Number of bytes in zData[] not counting final \000
  double rData;            // Floating-point value for this node
  int nRef;                // Number of cursors pointing to this node
  XmlTrace *pTrace;        // List of traces on this node
  XmlAttribute *pAttrList; // List of attributes at this node
  int flags;               // Various status bits
  short reset_counter : 1; // True if the next insert should reset the counter.
public:
  friend class XmlCursor;
};

#endif /* INTERFACE */

/*
** Return a pointer to the first subelement beneath this element.
*/
PRIVATE XmlNode *XmlNode::firstChild() const {
  return pChild ? pChild->leftmostSibling() : 0;
}

/*
** Return the index number for this element.
*/
PRIVATE int XmlNode::index() const {
  return iIndex;
}

/*
** Return the name of this element
*/
PRIVATE const char *XmlNode::name() const {
  return zName;
}

/*
** Return TRUE if this element is a leaf (if it contains no subelements)
** or return FALSE if this element does contain subelements.
*/
PRIVATE int XmlNode::is_leaf() const {
  return pChild==0;
}

/*
** Increase the number of outstanding references to this node by
** one.  The node might be unlinked from the XML tree but it will
** not be deleted until all references are released.
*/
PRIVATE void XmlNode::ref(){
  nRef++;
}

#if INTERFACE

/*
** This structure holds attribute information
*/
struct XmlAttribute {
  char *zName;            /* Name of this attribute */
  char *zValue;           /* Value of this attribute */
  XmlAttribute *pNext;    /* Pointer to the next attribute */
  ~XmlAttribute(){        /* Destructor (we must unlink the attribute from
                             pNext or we delete the rest of the list) */

    if( zName ){ delete [] zName; zName = 0; }
    if( zValue ){ delete [] zValue; zValue = 0; }
    if( pNext ){ delete pNext; pNext = 0; }
  }
};

/*
** object: XmlTrace
**
** A change to an XmlNode can provoke the execution of other functions.
** Each of these functions has an opportunity to overrule the change.
**
** There is an instance of the XmlTrace class for each callback associated 
** with an XmlNode.  The XmlTrace stores a pointer to the callback
** function with its arguments and information about when the callback
** should be invoked.
*/
class XmlTrace {
private:
  XmlNode *pNode;             // The node that holds the trace
  XmlTraceCallback xFunc;     // Function to call when the trace fires
  void (*xDel)(void*);        // Called to destroy pArg when deleted
  void *pArg;                 // Userdata for the xFunc() callback
  XmlTrace *pNext;            // Next trace on this same node
  int flags;                  // One or more flags.  See XML_TRACE_*
  int nActive;                // How many nested calls to this trace are
                              // active.
public:
  XmlTrace(){};
  ~XmlTrace(){ if( xDel ) xDel(pArg); };
  friend class XmlNode;
};

/*
** Traces invoke callback functions where the prototype of the
** callback function is as follows:
*/
typedef int (*XmlTraceCallback)(
  class XmlCursor *pCursor,    // Ptr to node where change occurred
  class XmlCursor *pReporter,  // Ptr to node holding the trace
  void *pUserData,             // User defined callback data
  int chngType                 // Type of change.
);

/*
** These are the allowed values for the chngType parameter
** of a trace callback.
*/
#define XML_TRACE_CREATE    1     // The XML element is newly created
#define XML_TRACE_MODIFY    2     // The XML element has been modified
#define XML_TRACE_DELETE    3     // The XML element has been deleted

/*
** The XML_TRACE_INDEX values of the chngType parameter are for internal use
** when a node is re-indexed.
**
** XML_TRACE_INDEX1 invokes deletion traces for the current node and for
** all its descendents. XML_TRACE_INDEX2 invokes creation traces for the
** current node and for all its descendents.
**
** Changing index is similar to the renaming operation implemented by the
** 'CURSOR mv' command, except that in the renaming operation the affected
** nodes are actually deleted and recreated. When changing index, nodes are not
** deleted or created, but we fire traces as if they were.
**
** Trace callbacks are not fired with either XML_TRACE_INDEX value as the
** chngType parameter.
*/
#define XML_TRACE_INDEX1    4     // The XML element is going to change index
#define XML_TRACE_INDEX2    5     // The XML element has changed index

/*
** Legal values of the XmlTrace.flags field that cannot be used
** in the third parameter to XmlCursor.trace().  These definitions
** are local because users of this library never need them.
** 
*/
#define XML_TRACE_MASK      7     // The user-settable bits
#define XML_TRACE_DEAD      8     // True if this trace is deleted
#define XML_TRACE_BUSY     16     // A callback for this trace is active

/*
** Legal values of the XmlNode.flags
*/
#define XML_NODE_TRACE      1     // True if executing traces
#define XML_NODE_DEAD       2     // True if this node has been deleted
#define XML_NODE_REAL       4     // True if the rData is valid
#define XML_NODE_STR        8     // True if zData is valid

#endif // INTERFACE

/*
** The following XmlNode is the root of the global Xml tree structure.
*/
static class XmlNode xmlRoot("/", 0);

#if INTERFACE

/*
** object: XmlCursor
**
** Each process implements a single global of XML data set formed from
** from a tree of XmlNode objects.  But the XmlNode objects may not
** be directly accessed by client code.  Instead, the client code must
** create one or more XmlCursor objects and use the XmlCursors to 
** access and/or modify the global Xml data tree.
**
** Each XmlCursor is a pointer to a single node in the XML data tree.
** The XmlCursor object supports methods to:<ol>
**
** <li> Return the basename or full pathname of the node to which the
**      cursor is currently pointing.
** 
** <li> Move the cursor to another nodes in the tree.  The node to
**      which the cursor is moved can be specified by either an
**      absolute or a relative pathname.
**
** <li> Query or change the data string associated with leaf nodes.
**
** <li> Create or delete children of interior nodes.
**
** <li> Add or remove callbacks (XmlTrace objects) that are invoked
**      when a node is accessed or modified.
** </ol>
*/
class XmlCursor {
  DEPENDS_ON(XmlNode)        // Tells makeheaders to define XmlNode first
private:
  class XmlNode *pNode;      // Xml node to which this cursor points
  int moveOk;                // TRUE if last move() worked ok.
  XmlAttribute *pAttr;       // Xml attribute to which this cursor points
  int attrMoveOk;            // TRUE if last attribute move() worked ok.

public:
  int operator==(const XmlCursor &other)const{ return pNode==other.pNode; };
  int operator!=(const XmlCursor &other)const{ return pNode!=other.pNode; };
  const char *name() const {return pNode->name();};
  const char *fullname() const {return pNode->fullname();};
  const char *basename() const {return pNode->basename();};
  const char *pwd() const {return pNode->pwd();};
  int index() const {return pNode->index();};
  int ok() const {return moveOk;};
  int attr_ok() const {return attrMoveOk;};
  void set(const char *zData, int nData = -1){ pNode->set(zData, nData);};
  void set(double rData){ pNode->set(rData); };
  const char *get()const{return pNode->get();};
  int length(){return pNode->length();};
  double get_real(){return pNode->get_real();};
  const char *coerce_real()const{return pNode->coerce_real();};
  int is_real() const {return pNode->is_real();};
  int is_int() const {return pNode->is_int();};
  int is_leaf() const {return pNode->is_leaf();};
  void ref(){ pNode->ref(); };
  int unref(){ return pNode->unref(); };
  int equal(XmlCursor left, XmlCursor right){
    return pNode->equal(left.pNode, right.pNode);
  }
  void trace(XmlTraceCallback xFunc, void *pArg, int flags,
             void(*xDel)(void*) = 0){
    pNode->addTrace(xFunc,pArg,flags,xDel);
  };
  void untrace(XmlTraceCallback xFunc, void *pArg, int flags){
    pNode->removeTrace(xFunc,pArg,flags);
  };
  int getTrace(int N, XmlTraceCallback *xFunc, void **pp, int *pFlags){
    return pNode->getTrace(N, xFunc, pp, pFlags);
  };
  void testDump() const {pNode->testDump(0);};
  int testRef() const {return pNode->nRef;};

  friend class XmlNode;
};

/*
** Legal values for the third argument to XmlCursor.trace() and
** XmlCursor.untrace().
** 
*/
#define XML_TRACE_SELF      1     // Apply traces to self
#define XML_TRACE_CHILDREN  2     // Apply traces to immediate children 
#define XML_TRACE_ALL       4     // Apply traces to all decendents

#endif // INTERFACE

/*
** Constructor for XmlNode
*/
PUBLIC XmlNode::XmlNode(const char *z, int n){
  pParent = 0;
  pUp = 0;
  pLeft = 0;
  pRight = 0;
  pChild = 0;
  pAttrList = 0;
  weight = 1;
  if( n<0 ) n = strlen(z);
  int i;
  for(i=n-2; i>0 && !(z[i]==':' && isdigit(z[i+1])); i--){}
  if( i>0 ){
    iIndex = atoi(&z[i+1]);
    n = i;
  }else{
    iIndex = -1;
  }
  zName = new char[n+1];
  if( zName ){
    strncpy(zName, z, n);
    zName[n] = 0;
  }
#if 0
  for(i=0; zName[i]; i++){
    if( isupper(zName[i]) ) zName[i] = tolower(zName[i]);
  }
#endif
  zFull = 0;
  zBase = 0;
  zData = 0;
  nData = 0;
  rData = 0.0;
  nRef = 0;
  pTrace = 0;
  flags = 0;
  reset_counter = 0;
  // printf("CREATE 0x%08x %s\n", (int)this, zName);
}

/*
** Destructor for XmlNode
*/
PUBLIC XmlNode::~XmlNode(){
  // printf("DELETE 0x%08x %s %s\n", (int)this, zName, zFull);
  removeAllTraces();
  pTrace = 0;
  if( zData ) delete [] zData;
  zData = 0;
  nData = 0;
  if( pChild ) pChild->die();
  pChild = 0;
  fratricide();
  pUp = 0;
  if( zFull ) delete [] zFull;
  zFull = zBase = 0;
  assert( zName!=0 );
  if( zName ) delete [] zName;
  zName = 0;
  if( pAttrList ) delete pAttrList;
  pAttrList = 0;
}

/*
** Figure out how much memory an XmlNode and its children have
** allocated.
**
** aMem[] is an array of integers that is initially zero.  Various
** integers in this array are increased by sizes and counts of
** XmlNode and its children.
**
**     aMem[0]     Total number of nodes
**     aMem[1]     Number of mallocs for node structure
**     aMem[2]     Total size of mallocs for node structure
**     aMem[3]     Number of mallocs for holding text data
**     aMem[4]     Total size of mallocs for holding text data
**     aMem[5]     Number of mallocs for holding full pathnames
**     aMem[6]     Total size of mallocs for holding full pathnames
**     aMem[7]     Number of mallocs for traces
**     aMem[8]     Total size of mallocs for holding traces
**     aMem[9]     Total number of attributes
**     aMem[10]    Total size of mallocs for holding attribute name and value
**
** The node itself and the nodes children are always included in these
** counts.  Siblings of this node are included only if the includeSiblings
** parameter is true.
*/
PRIVATE void XmlNode::memUsed(int *aMem, int includeSiblings){
  aMem[0]++;
  aMem[1]++;
  aMem[2] += sizeof(XmlNode);
  if( zName ){
    aMem[1]++;
    aMem[2] += strlen(zName) + 1;
  }
  if( zData ){
    aMem[3]++;
    aMem[4] += nData + 1;
  }
  if( zFull ){
    aMem[5]++;
    aMem[6] += strlen(zFull) + 1;
  }
  if( pTrace ){
    XmlTrace *p = pTrace;
    while( p ){
      aMem[7]++;
      aMem[8] += sizeof(XmlTrace);
      p = p->pNext;
    }
  }
  if( pAttrList ){
    XmlAttribute *pTemp = pAttrList;
    while( pTemp ) {
      aMem[9]++;
      aMem[10] += strlen(pTemp->zName) + 1;
      aMem[10] += strlen(pTemp->zValue) + 1;
      pTemp = pTemp->pNext;
    }
  }
  if( pChild ){
    pChild->memUsed(aMem, 1);
  }
  if( includeSiblings ){
    if( pLeft ) pLeft->memUsed(aMem, 1);
    if( pRight ) pRight->memUsed(aMem, 1);
  }
}

/*
** Remove all traces from this node.
*/
PRIVATE void XmlNode::removeAllTraces(){
  while( pTrace ){
    XmlTrace *pOld = pTrace;
    pTrace = pOld->pNext;
    delete pOld; pOld = 0;
  }
}

/*
** If nRef==0 (meaning there are not cursors pointing at this node)
** Then delete the node.  Otherwise, just delete all the data and
** mark the node for deletion when the last cursor reference
** goes away.
**
** This proc returns TRUE if it removes its XmlNode object.
*/
PRIVATE int XmlNode::die(){
  if( pParent==0 ) return 0;  // Never kill the root node.
  pParent = 0;
  if( nRef ){
  // printf("DIE 0x%08x %s %s\n", (int)this, zName, zFull);
    flags |= XML_NODE_DEAD;
    flags &= ~XML_NODE_REAL;
    if( zFull ){
      delete [] zFull;
      zFull = 0;
      zBase = 0;
    }
    if( zData ){
      delete [] zData;
      zData = 0;
    }
    nData = 0;
    if( pChild ){
      pChild->die();
      pChild = 0;
    }
    fratricide();
  }else{
    delete this;
    return 1;
  }
  return 0;
}

/*
** Unlink "this" from its sibling tree and from its parent.
** All children are also unlinked.  If there are no cursors
** pointing to this node, then the node is deleted.  If
** there are still cursors pointing to this need, the deletion
** is deferred until after the cursors move off of this node.
**
** This proc returns 1 if its object was removed as part of its operations.
*/
PRIVATE int XmlNode::unlink(){
  /*
  ** Kill off all children.
  */
  int rc = 0;
  ref();
  if( pChild ){
    pChild->unlinkAllSiblings();
    pChild = 0;
  }
  if( (flags & XML_NODE_DEAD)!=0 ){
    rc = unref();
    return rc;
  }
  flags |= XML_NODE_DEAD;

  /*
  ** Invoke the delete trace first before actually deleting
  ** the node.  This prevents the problem of calling the trace
  ** callback with a cursor that points to a deleted node.
  **
  ** The fact that delete traces happen before the fact make
  ** them different from create and modify traces which happen
  ** after the fact.
  */
  if( (flags & XML_NODE_TRACE)==0 ){
    XmlCursor cursor;
    cursor = this;
    flags |= XML_NODE_TRACE;
    int curInvalid = 0;
    if( invokeTrace(&cursor, 0, XML_TRACE_DELETE, &curInvalid) ){
      return 1; // this node has been removed by trace proc
    }
    flags &= ~XML_NODE_TRACE;
  }
  removeAllTraces();

  /*
  ** Unlink the node from the XML tree.
  */
  XmlNode **ppFrom;
  if( pUp ){
    ppFrom = pUp->pLeft==this ? &pUp->pLeft : &pUp->pRight;
  }else if( pParent==0 ){
    rc = unref();
    return rc;  // Never unlink the root node
  }else{
    ppFrom = &pParent->pChild;
  }
  if( pLeft==0 ){
    *ppFrom = pRight;
    if( pRight ) pRight->pUp = pUp;
    if( pUp ) pUp->balance();
  }else if( pRight==0 ){
    *ppFrom = pLeft;
    if( pLeft ) pLeft->pUp = pUp;
    if( pUp ) pUp->balance();
  }else if( pLeft->pRight==0 ){
    *ppFrom = pLeft;
    pLeft->pRight = pRight;
    pRight->pUp = pLeft;
    pLeft->pUp = pUp;
    pLeft->balance();
  }else if( pRight->pLeft==0 ){
    *ppFrom = pRight;
    pRight->pLeft = pLeft;
    pLeft->pUp = pRight;
    pRight->pUp = pUp;
    pRight->balance();
  }else{
    XmlNode *pLeftmost;
    for(pLeftmost=pRight; pLeftmost->pLeft; pLeftmost=pLeftmost->pLeft){}
    XmlNode *pOldUp = pLeftmost->pUp;
    assert( pOldUp->pLeft==pLeftmost );
    pOldUp->pLeft = pLeftmost->pRight;
    if( pOldUp->pLeft ){ pOldUp->pLeft->pUp = pOldUp; }
    pLeftmost->pUp = pUp;
    pLeftmost->pLeft = pLeft;
    pLeftmost->pRight = pRight;
    pLeft->pUp = pLeftmost;
    pRight->pUp = pLeftmost;
    *ppFrom = pLeftmost;
    pLeftmost->balance();
    pOldUp->balance();
  }
  pParent = 0;
  pUp = 0;
  pLeft = 0;
  pRight = 0;
  weight = 0;

  /*
  ** If there are cursors pointing to this node, then just set the
  ** XML_NODE_DEAD flag.  This will be a signal to the cursor to
  ** delete the node when it moves off.  If there are no cursors,
  ** then go ahead and delete this node right away.
  */
  if( die() ){
    return 1;
  }
  rc = unref();
  return rc;
}

/*
** Unlink "this" and all its siblings.
**
** This routine works like unlink() except that it does a whole
** tree of siblings all at once.  This saves the effort of rebalancing
** the sibling tree after every deletion.
**
** This proc returns TRUE if its object was removed as a result of its 
** operation.
*/
PRIVATE int XmlNode::unlinkAllSiblings(){
  /*
  ** It is convenenient to unlink siblings in order.  So go ahead
  ** and get the siblings to our left before doing ourself.
  */
  int rc = 0;
  ref();
  if( pLeft ){
    pLeft->unlinkAllSiblings();
    pLeft = 0;
  }
  if( (flags & XML_NODE_DEAD)!=0 ){
    rc = unref();
    return rc;
  }

  /*
  ** Get rid of all the children next.
  */
  if( pChild ){
    pChild->unlinkAllSiblings();
    pChild = 0;
  }
  if( (flags & XML_NODE_DEAD)!=0 ){
    rc = unref();
    return rc;
  }
  flags |= XML_NODE_DEAD;

  /*
  ** Invoke the delete trace first before actually deleting
  ** the node.  This prevents the problem of calling the trace
  ** callback with a cursor that points to a deleted node.
  **
  ** The fact that delete traces happen before the fact make
  ** them different from create and modify traces which happen
  ** after the fact.
  */
  if( (flags & XML_NODE_TRACE)==0 ){
    XmlCursor cursor;
    cursor = this;
    flags |= XML_NODE_TRACE;
    int curInvalid = 0;
    if( invokeTrace(&cursor, 0, XML_TRACE_DELETE, &curInvalid) ){
      return 1; // this node has been removed by trace proc
    }
    flags &= ~XML_NODE_TRACE;
  }
  removeAllTraces();

  /*
  ** Now get the siblings to the right.
  */
  if( pRight ){
    pRight->unlinkAllSiblings();
    pRight = 0;
  }

  /*
  ** If there are cursors pointing to this node, then just set the
  ** XML_NODE_DEAD flag.  This will be a signal to the cursor to
  ** delete the node when it moves off.  If there are no cursors,
  ** then go ahead and delete this node right away.
  */
  pUp = 0;
  weight = 0;
  if( die() ){
    return 1;
  }
  rc = unref();
  return rc;
}

/*
** Delete all siblings beneath "this" in the sibling btree.
*/
PRIVATE void XmlNode::fratricide(){
  if( pLeft ){ pLeft->die(); pLeft = 0; }
  if( pRight ){ pRight->die(); pRight = 0; }
  weight = 1;
}

/*
** Decrement the reference count on a node.
** Return TRUE if this proc actually deletes this object.
*/
PRIVATE int XmlNode::unref(){
  nRef--;
  assert( nRef>=0 );
  if( nRef==0 && (flags & XML_NODE_DEAD)!=0 ){
    delete this;
    return 1;
  }
  return 0;
}


/* 
** Return true iff this node is ancestral to pOther.  (This node is
** ancestral to itself.)
*/
PRIVATE int XmlNode::is_ancestor(const XmlNode *pOther) const {
  assert( pOther!=NULL );
  int depth_diff = pOther->depth()-depth();
  const XmlNode *p = pOther;
  for( int i=0;i<depth_diff;i++ ){
    assert( p->pParent );
    p = p->pParent;
  }
  return p==this;
}

/*
** Return the depth of this node.  The root node has depth 0.
*/
PRIVATE int XmlNode::depth() const {
  const XmlNode* p = this;
  int d=0;
  while( (p=p->pParent)!=NULL ){ d++; }
  return d;
}

/*
** Return TRUE iff the values of the referenced nodes are equal.
*/
PRIVATE int XmlNode::equal(XmlNode *pLeft, XmlNode *pRight){
  if( pLeft->is_real() && pRight->is_real() ){
    return pLeft->get_real()==pRight->get_real();
  }else{
    const char *z1, *z2;
    int n1, n2;
    z1 = pLeft->get();
    n1 = pLeft->length();
    z2 = pRight->get();
    n2 = pRight->length();
    return (z1==0 && z2==0) ||
               (z1!=0 && z2!=0 && n1==n2 && memcmp(z1,z2,n1)==0);
  }
}


#if 0 /* NOT USED */
/*
** Return true if the given string represents a floating-point number.
*/
PRIVATE STATIC int XmlNode::isFloat(const char *z){
  while( isspace(*z) ){ z++; }
  if( *z=='-' || *z=='+' ){ z++; }
  if( !isdigit(*z) ) return 0;
  z++;
  while( isdigit(*z) ){ z++; }
  if( *z=='.' ){
    z++;
    if( !isdigit(*z) ) return 0;
    z++;
    while( isdigit(*z) ){ z++; }
  }
  if( *z=='e' || *z=='E' ){
    z++;
    if( *z=='+' || *z=='-' ){ z++; }
    if( !isdigit(*z) ) return 0;
    z++;
    while( isdigit(*z) ){ z++; }
  }
  while( isspace(*z) ){ z++; }
  return *z==0;
}
#endif

#if 0 /* NOT USED */
/*
** *z is an ascii representation of a floating-point number.  Truncate
** the precision of that number to be no more than 13 digits.
*/
PRIVATE STATIC void XmlNode::truncatePrecision(char *z){
  int i, j;
  int n;

  for(i=j=n=0; z[i]; i++){
    if( isdigit(z[i]) ){
      if( n<13 ) z[j++] = z[i];
      n++;
    }else{
      if( z[i]=='e' || z[i]=='E' ) n = 13;
      z[j++] = z[i];
    }
  }
  z[j] = 0;
}
#endif

/*
** Change the CDATA content of a node to a copy of zNewData.  The new CDATA
** is stored in memory obtained from the "new" operator.  Any prior CDATA
** is deleted.  If zNewData==0 or nNewData==0 then the prior content 
** is deleted but no new data is added.  If nNewData<0 then strlen(zNewData)
** is used instead.
*/
PRIVATE void XmlNode::set(const char *zNewData, int nNewData){
  if( pChild ){
    pChild->unlinkAllSiblings();
    pChild = 0;
  }
  if( nNewData<0 && zNewData!=0 ) nNewData = strlen(zNewData);
  if( zNewData==0 || nNewData==0 ){
    if( zData==0 || nData==0 ) return;
    if( zData ) delete [] zData;
    zData = 0;
    nData = 0;
    flags &= ~XML_NODE_REAL;
  }else{
    if( (flags & XML_NODE_REAL)==0 && zData && nNewData==nData && 
        memcmp(zNewData, zData, nData)==0 ) return;
    if( zData ) delete [] zData;
    nData = nNewData;
    zData = new char[nData+1];
    if( zData ){
      memcpy(zData, zNewData, nData);
      zData[nData] = 0;
    }else{
      nData = 0;
    }
    flags &= ~XML_NODE_REAL;
    
  }
  if( (flags & XML_NODE_TRACE)==0 ){
    XmlCursor cursor;
    cursor = this;
    flags |= XML_NODE_TRACE;
    int curInvalid = 0;
    if( invokeTrace(&cursor, 0, XML_TRACE_MODIFY, &curInvalid) ){
      return; // this node has been removed by trace proc
    }
    flags &= ~XML_NODE_TRACE;
  }
}

/*
** Change the CDATA content of a node to the floating-point
** value rNewData.
*/
PRIVATE void XmlNode::set(double rNewData){

  /* Determine if the change is needed.  We use the same criterion
  ** for equality that the Compare() and Diff() functions do.
  */
  if( (flags & XML_NODE_REAL)!=0 && rNewData==rData ) return;
  XmlNode *pNew = new XmlNode("set_test", -1);
  pNew->set(to_string(rNewData), -1);
  int same = equal(pNew, this);
  delete pNew;
  if( same ) return;

  flags |= XML_NODE_REAL;
  rData = rNewData;
  if( zData ){
    delete [] zData;
    zData = 0;
    nData = 0;
  }
  if( pChild ){
    pChild->unlinkAllSiblings();
    pChild = 0;
  }
  if( (flags & XML_NODE_TRACE)==0 ){
    XmlCursor cursor;
    cursor = this;
    flags |= XML_NODE_TRACE;
    int curInvalid = 0;
    if( invokeTrace(&cursor, 0, XML_TRACE_MODIFY, &curInvalid) ){
      return; // this node has been removed by trace proc
    }
    flags &= ~XML_NODE_TRACE;
  }
}

/*
** Return TRUE if the node holds floating-point data.
*/
PRIVATE int XmlNode::is_real() const {
  return (flags & XML_NODE_REAL)!=0;
}

/*
** Return TRUE if the node holds an integer value.
*/
PRIVATE int XmlNode::is_int() const {
  return (flags & XML_NODE_REAL)!=0 && rData==floor(rData);
}

/*
** Return the floating point value of the node.
**
** SIDE EFFECT: This routine actually changes the type of the
** node to XML_NODE_REAL if it isn't currently of that type.
*/
PRIVATE double XmlNode::get_real(){
  if( (flags & XML_NODE_REAL)==0 ){
    rData = zData ? atof(zData) : 0.0;
    flags |= XML_NODE_REAL;
  }
  return rData;
}

/*
** Return the value of the node, if the node contains a floating point value.
** If it doesn't, set the value of the node to 0.0 and return that.
*/
PRIVATE char *XmlNode::coerce_real(){
  int bIsReal = 1;
  if( zData==0 ){
    if( (flags & XML_NODE_REAL)!=0 ){
      stringify();
      return zData;
    }
    bIsReal = 0;
  }else{
    char *zEnd;
    errno = 0;
    strtod(zData, &zEnd);
    if( errno==ERANGE || zData==zEnd ){
      bIsReal = 0;
    }else{
      while( *zEnd && isspace(*zEnd) ) ++zEnd;
      if( *zEnd ){
        bIsReal = 0;
      }
    }
  }
  if( !bIsReal ){
    if( zData ) delete [] zData;
    zData = new char[4];
    if( zData ) strcpy(zData, "0.0");
    nData = 3;
  }
  return zData;
}

/*
** Return the string value of the node.
*/
PRIVATE const char *XmlNode::get(){
  if( zData==0 && (flags & XML_NODE_REAL)!=0 ) stringify();
  return zData;
}

/*
** Return the string value of the node.
*/
PRIVATE int XmlNode::length(){
  if( zData==0 && (flags & XML_NODE_REAL)!=0 ) stringify();
  return nData;
}

/*
** Give the node a valid string representation
*/
PRIVATE void XmlNode::stringify(){
  char *zBuf = to_string(rData);
  nData = strlen(zBuf);
  zData = new char[nData+1];
  memcpy(zData, zBuf, nData+1);
}

/*
** Convert the floating point data into a string.
** The return is a static buffer; the caller must use it
** immediately after calling this routine so that other
** callers won't interfere.
*/
PRIVATE char *XmlNode::to_string(double data){
  static char zBuf[50];
  sprintf(zBuf,"%.12g",data);
  return zBuf;
}

/*
** Decode the XML escape characters in the this->zData field.
** "&lt;" is converted into "<", "&gt;" is changed to ">",
** "&amp;" is changed to "&", and "&#DDD;" is changed to a character
** whose numeric value is DDD.  The conversion is done in-place.
*/
PRIVATE void XmlNode::decodeXml(){
  decodeXml(zData);
}
PRIVATE STATIC size_t XmlNode::decodeXmlString(char *zText){
  int i, j;
  for(i=j=0; zText[i]; i++){
    if( zText[i]!='&' ){
      zText[j++] = zText[i];
    }else if( zText[i+1]=='l' && zText[i+2]=='t' && !isalnum(zText[i+3]) ){
      zText[j++] = '<';
      i += 2 + (zText[i+3]==';');
    }else if( zText[i+1]=='g' && zText[i+2]=='t' && !isalnum(zText[i+3]) ){
      zText[j++] = '>';
      i += 2 + (zText[i+3]==';');
    }else if( zText[i+1]=='a' && zText[i+2]=='m' && zText[i+3]=='p'
             && !isalnum(zText[i+4]) ){
      zText[j++] = '&';
      i += 3 + (zText[i+4]==';');
    }else if( zText[i+1]=='q' && strncmp(&zText[i+1],"quot",4)==0
             && !isalnum(zText[i+5]) ){
      zText[j++] = '"';
      i += 4 + (zText[i+5]==';');
    }else if( zText[i+1]=='a' && strncmp(&zText[i+1],"apos",4)==0
             && !isalnum(zText[i+5]) ){
      zText[j++] = '\'';
      i += 4 + (zText[i+5]==';');
    }else if( zText[i+1]=='#' && isdigit(zText[i+2]) ){
      int k = 1;
      while( isdigit(zText[i+k+2]) ){ k++; }
      zText[j++] = atoi(&zText[i+2]);
      i += k + 1 + (zText[i+k+2]==';');
    }else{
      zText[j++] = zText[i];
    }
  }
  zText[j] = 0;
  return j;
}

PRIVATE void XmlNode::decodeXml(char *zText){
  nData = XmlNode::decodeXmlString(zText);
}

/*
** Invoke traces on "this" and on all ancestor XmlNodes of this.
** The "level" parameter determines what kinds of traces are invoked:
**
**       0             XML_TRACE_SELF and XML_TRACE_ALL
**       1             XML_TRACE_CHILDREN and XML_TRACE_ALL
**       2 or more     XML_TRACE_ALL
**
** This routine automatically calls itself on ancestor nodes, incrementing
** the "level" value with each step of the ancestor tree.
**
** This routine returns TRUE if its XmlNode has been deleted (presumably
** because of an executed trace).  It also returns a value in pCurInvalid
** if it is pointing to an XmlNode that has been deleted as a result
** of this operation.
*/
PRIVATE int XmlNode::invokeTrace(
  XmlCursor *pCur,   // Pointer to node which changed
  int level,         // See comment above
  int chngType,      // Type of change
  int *pCurInvalid   // Returns TRUE if node that pCur references is invalid
){
  return invokeTrace(pCur, level, chngType, pCurInvalid, 0);
}
PRIVATE int XmlNode::invokeTrace(
  XmlCursor *pCur,   // Pointer to node which changed
  int level,         // See comment above
  int chngType,      // Type of change
  int *pCurInvalid,  // Returns TRUE if node that pCur references is invalid
  int callLevel      // Level of recursion in invokeTrace
){
  int mask;
  switch( level ){
    case 0:   mask = XML_TRACE_SELF | XML_TRACE_ALL;       break;
    case 1:   mask = XML_TRACE_CHILDREN | XML_TRACE_ALL;   break;
    default:  mask = XML_TRACE_ALL;                        break;
  }
  XmlCursor cur = this;  // Cursor pointing to reporting XmlNode
  *pCurInvalid = 0;

  /* This node may have been previously deleted.  If that's the case,
  ** the fullname will be null or blank, and the trace callbacks will not
  ** be useful (and potentially harmful).  So don't run them.
  */
  const char *zModFull = pCur->fullname();
  const char *zCurFull = fullname();
  if( !zModFull || !*zModFull ) return 0;
  if( !zCurFull || !*zCurFull ) return 0;

  /* Don't allow a "real" delete on this node while invoking traces
  */
  ref();

  /* Don't allow the node referenced by pCur to be truly deleted while
  ** invoking traces.
  */
  assert(pCur->pNode);
  pCur->ref();

  /* If we're changing our index, we need to invoke traces on our children.
  */
  if( pChild && (chngType==XML_TRACE_INDEX1 || chngType==XML_TRACE_INDEX2) ){
    XmlNode *pNextNode = 0;
    for( XmlNode* pNode=pChild->leftmostSibling(); pNode; pNode=pNextNode ){
      pNextNode = pNode->next();
      XmlCursor cur2(pNode);
      if( (pNode->flags & XML_NODE_TRACE)==0 ){
        pNode->flags |= XML_NODE_TRACE;
        int curInvalid = 0;
        if( pNode->invokeTrace(&cur2, 0, chngType, &curInvalid, callLevel+1) ){
          pNode = 0;
        }else{
          pNode->flags &= ~XML_NODE_TRACE;
        }
      }
    }
  }

  XmlTrace *p, *pNext;
  for(p=pTrace; p!=0; p=pNext){
    if( p->flags & mask ){
      if( p->nActive>0 ){
        pNext = p->pNext;
        continue;
      }
      p->nActive++;

      int irc = 0;
      if( chngType==XML_TRACE_INDEX1 ){
        irc = p->xFunc(pCur, &cur, p->pArg, XML_TRACE_DELETE);
      }else if( chngType==XML_TRACE_INDEX2 ){
        irc = p->xFunc(pCur, &cur, p->pArg, XML_TRACE_CREATE);
      }else{
        irc = p->xFunc(pCur, &cur, p->pArg, chngType);
      }

      /* As a result of the trace function above, pCur's and cur's nodes should
      ** still exist, but may be pending-deletion.  So don't check on fullnames.
      ** Original code here:
      **
      ** assert( pCur->fullname() && pCur->fullname()[0] );
      ** assert( cur.fullname() && cur.fullname()[0] );
      */

      /* If the trace function fails (irc==TRUE) then remove the trace 
      ** function.
      */
      if( irc ){
        removeTrace(p->xFunc, p->pArg, p->flags & XML_TRACE_MASK);
      }
      p->nActive--;
    }
    pNext = p->pNext;

    /* It is OK to delete p without removing the reference to it on the
    ** trace list.  The reference has already been removed in removeTrace().
    */
    if( p->nActive==0 && (p->flags & XML_TRACE_DEAD) ){
      delete p; p = 0;
    }
  }
  if( pParent ){
    if( chngType==XML_TRACE_INDEX1 ){ chngType = XML_TRACE_DELETE; }
    if( chngType==XML_TRACE_INDEX2 ){ chngType = XML_TRACE_CREATE; }
    if( (pParent->flags & XML_NODE_TRACE)==0 ){
      pParent->flags |= XML_NODE_TRACE;
      int curInvalid = 0;
      if( pParent->invokeTrace(pCur, level+1, chngType, &curInvalid, 
          callLevel+1) 
      ){
        pParent = 0;
      }
      if( pParent ){
        pParent->flags &= ~XML_NODE_TRACE;
      }
      assert(curInvalid==0);
    }
  }
#if 0
  /* This *really* slows things down but can be useful -- it potentially 
  ** gives you a core file with a more informative call stack.
  */
  if( callLevel==0 ){ XmlHash::StrictCheckAll(chngType); }
#endif

  /* If a trace deleted the node referenced by pCur, inform the caller
  */
  *pCurInvalid = pCur->unref();

  /* This will return TRUE if 'this' no longer exists
  */
  int rc = unref();
  return rc;
}

/*
** Add a new trace to the end of the trace list.
**
** If a new trace is added by a trace callback on the same node,
** then the new trace will normally be executed after the current
** trace returns.  Except if the current trace is deleted and also
** happens to be the last trace on the trace list before the new
** one is added, then the new trace does not get executed.  Note
** that this complication only comes into play when you try to 
** delete and create traces from within trace callbacks.
*/
PRIVATE void XmlNode::addTrace(
  XmlTraceCallback xCallback,      // The callback routine
  void *pArg,                      // User argument to the callback
  int flags,                       // Flags that determine which trace fires
  void (*xDel)(void*)              // Called to destroy pArg when trace deleted
){
  XmlTrace *pNew = new XmlTrace;
  pNew->pNode = this;
  pNew->xFunc = xCallback;
  pNew->xDel = xDel;
  pNew->pArg = pArg;
  pNew->pNext = 0;
  pNew->flags = flags & (XML_TRACE_SELF | XML_TRACE_CHILDREN | XML_TRACE_ALL);
  pNew->nActive = 0;
  if( pTrace==0 ){
    pTrace = pNew;
  }else{
    XmlTrace *p;
    for(p = pTrace; p->pNext; p = p->pNext){}
    p->pNext = pNew;
  }
}

/*
** Remove a trace from the trace list.
**
** It is OK to delete a trace from within a trace callback.  When that
** happens, the XmlTrace structure is not actually deleted.  Instead,
** the XML_TRACE_DEAD flag is set and the XmlTrace structure gets
** deleted when the callback returns.
*/
PRIVATE void XmlNode::removeTrace(
  XmlTraceCallback xCallback,
  void *pArg,
  int flags
){
  XmlTrace **ppPrior = &pTrace;
  XmlTrace *p = 0;
  flags &= XML_TRACE_MASK;
  while( (p=*ppPrior)!=0 ){
    if( p->xFunc==xCallback && p->pArg==pArg && 
      flags==(p->flags & ~XML_TRACE_BUSY)
    ){
      *ppPrior = p->pNext;
      if( p->nActive>0 ){
        p->flags |= XML_TRACE_DEAD;
      }else{
        delete p; p = 0;
      }
      break;
    }else{
      ppPrior = &p->pNext;
    }
  }
}

/*
** Return information about the N-th trace.
*/
PRIVATE int XmlNode::getTrace(
  int N,                       // Which trace are we interested in
  XmlTraceCallback *pxFunc,    // The callback function
  void **ppArg,                // The user data
  int *pFlags                  // Flags
){
  XmlTrace *p;
  for(p=pTrace; N>0 && p; p=p->pNext){ N--; }
  if( p ){
    *pxFunc = p->xFunc;
    *ppArg = p->pArg;
    *pFlags = p->flags & XML_TRACE_MASK;
    return 1;
  }
  return 0;
}

/*
** Get the full pathname of the parent element.  If this is the root
** element, return "/".
*/
PRIVATE const char *XmlNode::pwd() const {
  if( pParent==0 ) return "/";
  return pParent->fullname();
}

/*
** Return the full pathname of this node.
**
** The pathname may have been previously computed and stored
** in this->zFull.  If so, just return it.  If not, then compute
** it this time and store it in this->zFull in case this routine
** is called again.
*/
PRIVATE const char *XmlNode::fullname(){
  if( zFull ){ return zFull; }
  if( pParent==0 ){
    return (flags & XML_NODE_DEAD)==0 ? "/" : "";
  }
  const char *zPwd = pParent->fullname();
  if( zPwd[1]==0 ){ zPwd = ""; }
  int nPwd = strlen(zPwd);
  char zIdx[30];
  sprintf(zIdx,"%d",iIndex);
  zFull = new char[nPwd + strlen(zName) + strlen(zIdx) + 3];
  if( zFull==0 ) return 0;
  sprintf(zFull,"%s/%s:%s",zPwd,zName,zIdx);
  zBase = &zFull[nPwd+1];
  return zFull;
}

/*
** Return the "basename".  The basename is the element name with :index
** appended.
*/
PRIVATE const char *XmlNode::basename(){ 
  if( zBase==0 ){
    fullname();
  }
  return zBase;
}

/*
** Locate the sibling of this (or this itself) that has XmlNode.zName==zTag
** and XmlNode.iIndex==idx.  Only siblings beneath *this are searched.  
** XmlNode.zName must match zTag exactly or NULL is returned.  But if 
** XmlNode.iIndex!=idx then the node with the smallest index greater than 
** or equal to idx is returned.  If all nodes with the same XmlNode.zName
** have an index smaller then idx, then the node with the largest index
** is returned.
*/
PRIVATE XmlNode *XmlNode::findSameName(const char *zTag, int n, int idx){
  XmlNode *pThis = this;
  XmlNode *pMatch = 0;
  if( n==0 ) n = strlen(zTag);
  while( pThis ){
    int c = strncmp(pThis->zName, zTag, n);
    if( c==0 ){
      if( pThis->zName[n]!=0 ){
        c = 1;
      }else{
        c = pThis->iIndex - idx;
        if( c==0 ){
          pMatch = pThis;
          break;
        }
        if( pMatch==0 || (c<0 && pMatch->iIndex>pThis->iIndex) ){
          pMatch = pThis;
        }
      }
    }
    if( c<0 ){
      pThis = pThis->pRight;
    }else{
      pThis = pThis->pLeft;
    }
  }
  return pMatch;
}

/*
** zPath is a pathname of an XML element.  If zPath begins with "/"
** then it is an absolute pathname.  Otherwise, zPath is interpreted
** relative to *this.  The job of this routine is to walk down zPath,
** locate the node it names, and return a pointer to that node.
*/
PRIVATE XmlNode *XmlNode::findByPath(const char *zPath){
  if( !zPath ) { return 0; }
  XmlNode *p;
  if( zPath[0]=='/' ){
    while( zPath[0]=='/' ) ++zPath;
    p = &xmlRoot;
  }else{
    p = this;
  }
  while( zPath[0] ){
    int idx;
    int i, j;
    long lIdx;
    char* zEnd;
    i=-1;
    do {
        for(++i; zPath[i] && zPath[i]!=':' && zPath[i]!='/'; ++i) ;
        if( zPath[i]!=':' || !zPath[i+1] ){ idx=0; break; }
        lIdx = strtol(&zPath[i+1], &zEnd, 10);
        idx = (int)lIdx;
    } while( lIdx==LONG_MIN || lIdx==LONG_MAX || zEnd==&zPath[i+1] );
    if( zPath[i]==':' ){
      idx = atoi(&zPath[i+1]);
      for(j=i+1; zPath[j] && zPath[j]!='/'; ++j){}
    }else{
      j = i;
      idx = 0;
    }
    if( i==2 && strncmp(zPath,"..",2)==0 ){
      if( p->pParent ) p = p->pParent;
      zPath += j;
      while( zPath[0]=='/' ) zPath++;
      continue;
    }
    if( i==1 && zPath[0]=='.' ){
      zPath += j;
      while( zPath[0]=='/' ) zPath++;
      continue;
    }
    if( p->pChild==0 ){
      p = 0;
      break;
    }
    p = p->pChild->findSameName(zPath, i, idx);
    if( p==0 || (p->iIndex!=idx && i!=j) ){
      p = 0;
      break;
    }
    zPath += j;
    while( zPath[0]=='/' ) zPath++;
  }
  return p;
}

/*
** Insert "this" as a child of pParent.  Return the
** number of errors.  Insertion can fail as a result of the invocation of
** a trace routine.
**
** If this->iIndex is initially less than zero then it will be set
** to one more than the largest existing entry with the same zName.
** If this->iIndex is initially non-negative, then an attempt is made
** to create a new entry with the given index.  But if there is a 
** collision, this->iIndex is set to -1 and the process starts over.
**
** For example, suppose the tree holds these values: A:0 B:0 B:2 B:3 C:1
** An attempt to insert B:1 succeeds and the index is unchanged.
** But an attempt to insert B:2 would collide with an existing entry
** and so the index would be changed from 2 to 4.
*/
PRIVATE int XmlNode::insert(XmlNode *_pParent){
  static int nextIndex = 0;
  if( reset_counter ){  // FIXME: hack
    nextIndex = 0;
    reset_counter = 0;
    return 0;
  }

  /*
  ** Initialize the pointers of the node being inserted
  */
  pParent = _pParent;
  pLeft = 0;
  pRight = 0;
  weight = 1;

  /* If the node index is unspecified, create one.
  */
  if( iIndex<0 ){
    if( nextIndex==2147483647 ){
      iIndex = rand();
    }else{
      iIndex = nextIndex++;
    }
  }

  /* Insert the node into its parent's child-tree
  */
  XmlNode *pNode = pParent->pChild;
  if( pNode==0 ){
    /* Simple special case: the parent has no children..
    */
    pParent->pChild = this;
    pUp = 0;
  }else{
    /* More complex.  We have to figure where in the b-tree
    ** of siblings to put the new child.
    */
    int nName = strlen(zName);
    while( pNode ){
      int c = strncmp(zName, pNode->zName, nName);
      if( c==0 ){
        if( pNode->zName[nName]!=0 ){
          c = -1;
        }else if( iIndex>=0 ){
          c = iIndex - pNode->iIndex;
          if( c==0 ){
            iIndex = -1;
            pNode = pParent->pChild;
            continue;
          }
        }
      }
      if( c<0 ){
        if( pNode->pLeft==0 ){
          pNode->pLeft = this;
          break;
        }
        pNode = pNode->pLeft;
      }else{
        if( pNode->pRight==0 ){
          pNode->pRight = this;
          break;
        }
        pNode = pNode->pRight;
      }
    }
    assert( pNode!=0 );
    pUp = pNode;
    if( iIndex<0 ){
      XmlNode *pPrev = prev();
      if( pPrev && strcmp(pPrev->zName,zName)==0 ){
        iIndex = pPrev->iIndex+1;
      }else{
        iIndex = 0;
      }
    }
    pNode->balance();
  }

  /* Clear any data associated with the parent.  Make sure to call
  ** modification callbacks if we do so.
  */
  if( pParent->zData ){
    delete [] pParent->zData;
    pParent->zData = 0;
    pParent->nData = 0;

    if( (pParent->flags & XML_NODE_TRACE)==0 ){
      XmlCursor cur(pParent);
      pParent->flags |= XML_NODE_TRACE;
      int curInvalid = 0;
      if( pParent->invokeTrace(&cur, 0, XML_TRACE_MODIFY, &curInvalid) ){
        pParent = 0;  // pParent was removed by trace proc
      }
      if( pParent ){
        pParent->flags &= ~XML_NODE_TRACE;
      }
    }
  }

  /* After constructing the node, activate creation traces on 
  ** ancestor nodes.  Doing this trace can cause this newly created
  ** node to fail.
  */
  int curInvalid = 0;
  if( pParent && (pParent->flags & XML_NODE_TRACE)==0 ){
    XmlCursor cursor;
    cursor = this;
    pParent->flags |= XML_NODE_TRACE;
    if( pParent->invokeTrace(&cursor, 1, XML_TRACE_CREATE, &curInvalid) ){
      pParent = 0;  // pParent was removed by trace proc
    }
    if( pParent ){
      pParent->flags &= ~XML_NODE_TRACE;
    }
  }
  return curInvalid;  // If TRUE, this node was removed by a trace routine
}

/*
** Create an attribute and add it to this node
*/
PRIVATE void XmlNode::createAttribute(const char *name, const char *value){
  XmlAttribute *pNew = new XmlAttribute;
  XmlAttribute *pTemp = pAttrList;
  XmlCursor cursor;
  cursor = this;

  pNew->zName = new char[strlen(name)+1];
  strcpy(pNew->zName, name);
  pNew->zValue = new char[strlen(value)+1];
  strcpy(pNew->zValue, value);
  pNew->pNext = 0;
  if( pTemp ){
    while( pTemp->pNext!=0 ) pTemp=pTemp->pNext;
    pTemp->pNext = pNew;
  }else{
    pAttrList = pNew;
  }
  if( (flags & XML_NODE_TRACE)==0 ){
    flags |= XML_NODE_TRACE;
    int curInvalid = 0;
    if( invokeTrace(&cursor, 0, XML_TRACE_MODIFY, &curInvalid)==0 ){
      flags &= ~XML_NODE_TRACE;
    }
  }
}

/*
** Change an attribute's value to a new value
*/
PRIVATE void XmlNode::changeAttribute(XmlAttribute *oldAttr, const char *value){
  XmlAttribute *pAttr = oldAttr;
  XmlCursor cursor;
  cursor = this;

  if( !pAttr->zValue && !value ){ return; }
  if( pAttr->zValue && value && strcmp(pAttr->zValue, value)==0 ){ return; }

  if( pAttr->zValue ) delete [] pAttr->zValue;
  pAttr->zValue = new char[strlen(value)+1];
  strcpy(pAttr->zValue, value);
  if( (flags & XML_NODE_TRACE)==0 ){
    flags |= XML_NODE_TRACE;
    int curInvalid = 0;
    if( invokeTrace(&cursor, 0, XML_TRACE_MODIFY, &curInvalid)==0 ){
      flags &= ~XML_NODE_TRACE;
    }
  }
}

/*
** Balance "this" node and all parent nodes back up to the root of
** the sibling tree.
*/
PRIVATE void XmlNode::balance(){
  XmlNode *p = this;
  XmlNode **pp;
  while( p ){
    if( p->pUp ){
      pp = &p->pUp->pRight;
      if( p != *pp ){
        pp = &p->pUp->pLeft;
      }
    }else{
      assert( p->pParent!=0 );
      pp = &p->pParent->pChild;
    }
    assert( p == *pp );
    XmlNode::balanceOneLayer(pp);
    p = *pp;
    if( p->pUp==0 && p->pParent ){
      p->pParent->pChild = p;
    }
    p = p->pUp;
  }
}

/*
** Balance the subtree rooted at *ppTop.
*/
PRIVATE void XmlNode::balanceOneLayer(XmlNode **ppTop){
  XmlNode *n;      // The value of *ppTop
  int l,r;         // Weight of left and right children
  int a,b;         // Weights of various nodes

  if( ppTop==0 || (n=*ppTop)==0 ) return;
  l = n->pLeft ? n->pLeft->weight: 0;
  r = n->pRight ? n->pRight->weight: 0;
  if( l>r*2 ){    /* Too heave on the left side */
    XmlNode *nl;     /* Pointer to left child */
    XmlNode *nr;     /* Pointer to right child */
    int ll, lr;       /* Weight of left child's left and right child */
    nl = n->pLeft;
    ll = nl->pLeft ? nl->pLeft->weight: 0;
    lr = nl->pRight ? nl->pRight->weight: 0;
    if( ll>lr || nl->pRight==0 ){
      /*
      ** Convert from:  n     to:  nl
      **               / \        /  \
      **              nl  c      a    n
      **             /  \            / \
      **            a    b          b   c
      */
      n->pLeft = nl->pRight;
      if( n->pLeft ){ n->pLeft->pUp = n; }
      nl->pRight = n;
      nl->pUp = n->pUp;
      n->pUp = nl;
      n->weight = a = r + lr + 1;
      nl->weight = a + ll + 1;
      *ppTop = nl;
    }else{
      /*
      ** Convert from:  n        to:  nr
      **               / \           /  \
      **             nl   d        nl    n
      **            /  \          / \   / \
      **           a    nr       a   b c   d
      **               /  \
      **              b    c
      */
      int lrl, lrr;    /* Weight of Great-grandchild nodes */
      nr = nl->pRight;
      lrl = nr->pLeft ? nr->pLeft->weight: 0;
      lrr = nr->pRight ? nr->pRight->weight: 0;
      nl->pRight = nr->pLeft;
      nr->pLeft = nl;
      n->pLeft = nr->pRight;
      nr->pRight = n;
      nr->pUp = n->pUp;
      nl->pUp = nr;
      n->pUp = nr;
      if( nl->pRight ){ nl->pRight->pUp = nl; }
      if( n->pLeft ){ n->pLeft->pUp = n; }
      n->weight = a = lrr + r + 1;
      nl->weight = b = ll + lrl + 1;
      nr->weight = a + b + 1;
      *ppTop = nr;
    }
  }else if( r>l*2 ){/* Too deep on the right side */
    XmlNode *nl;     /* Pointer to left child */
    XmlNode *nr;     /* Pointer to right child */
    int rl, rr;       /* Weight of right child's left and right child */
    nr = n->pRight;
    rl = nr->pLeft ? nr->pLeft->weight: 0;
    rr = nr->pRight ? nr->pRight->weight: 0;
    if( rr>rl || nr->pLeft==0 ){
      /*
      ** Convert from:  n         to:  nr
      **               / \            /  \
      **              a   nr         n    c 
      **                 /  \       / \
      **                b    c     a   b
      */
      n->pRight = nr->pLeft;
      nr->pLeft = n;
      nr->pUp = n->pUp;
      n->pUp = nr;
      if( n->pRight ){ n->pRight->pUp = n; }
      n->weight = a = l + rl + 1;
      nr->weight = a + rr + 1;
      *ppTop = nr;
    }else{
      /*
      ** Convert from:  n         to:  nl
      **               / \            /  \
      **              a   nr         n    nr
      **                 /  \       / \   / \
      **               nl    d     a   b c   d
      **              /  \
      **             b    c
      */
      int rll,rlr;    /* Weights of great-grandchild nodes */
      nl = nr->pLeft;
      rll = nl->pLeft ? nl->pLeft->weight: 0;
      rlr = nl->pRight ? nl->pRight->weight: 0;
      nr->pLeft = nl->pRight;
      nl->pRight = nr;
      n->pRight = nl->pLeft;
      nl->pLeft = n;
      nl->pUp = n->pUp;
      n->pUp = nl;
      nr->pUp = nl;
      if( n->pRight ){ n->pRight->pUp = n; }
      if( nr->pLeft ){ nr->pLeft->pUp = nr; }
      n->weight = a = l + rll + 1;
      nr->weight = b = rr + rlr + 1;
      nl->weight = a + b + 1;
      *ppTop = nl;
    }
  }else{ /* Node is already balanced.  Just recompute its weight. */
    n->weight = l + r + 1;
  }
}

/*
** Return a pointer to the next sibling.  Return NULL if this is the
** first sibling.
*/
PRIVATE XmlNode *XmlNode::next(){
  XmlNode *p = this;
  if( p->pRight==0 ){
    while( p->pUp && p->pUp->pRight==p ){
      p = p->pUp;
    }
    p = p->pUp;
  }else{
    p = p->pRight;
    while( p->pLeft ){
      p = p->pLeft;
    }
  }
  return p;
}

/*
** Return a pointer to the previous sibling.  Return NULL if this is
** the first sibling.
*/
PRIVATE XmlNode *XmlNode::prev(){
  XmlNode *p = this;
  if( p->pLeft==0 ){
    while( p->pUp && p->pUp->pLeft==p ){
      p = p->pUp;
    }
    p = p->pUp;
  }else{
    p = p->pLeft;
    while( p->pRight ){
      p = p->pRight;
    }
  }
  return p;
}

/*
** Recursively print all information in the tree of XML nodes.
*/
PRIVATE void XmlNode::testDump(int indent) const {
  if( pLeft ) pLeft->testDump(indent);
  printf("%*s%p: p=%p u=%p l=%p r=%p c=%p %s\n",
     indent, "", this, pParent, pUp, pLeft, pRight, pChild,
     zBase ? zBase : zName);
  if( pChild ) pChild->testDump(indent+2);
  if( pRight ) pRight->testDump(indent);
}

/*
** Return a pointer to the left-most sibling of *this.
*/
PRIVATE XmlNode *XmlNode::leftmostSibling(){
  XmlNode *p = this;
  while( p->pLeft ){ p = p->pLeft; }
  return p;
}

/*
** Erase the cached basename and fullname of a node and of
** all its descendents.
**
** The correct operation of this routine depends upon the following invariant:
** If a node's cache is already erased, then all of its descendents 
** also have an erased cache.
*/
PRIVATE void XmlNode::eraseCachedNames(){
  if( !zBase ) return;
  eraseOneCachedName();
  if( !pChild ) return;

  for(XmlNode* pNode=pChild->leftmostSibling(); pNode; pNode=pNode->next()){
    if( pNode->zBase ){
      pNode->eraseCachedNames();
    }
  }
}

/*
** Erase the cached basename and fullname of a node.
*/
PRIVATE void XmlNode::eraseOneCachedName(){
  if( zBase ){
    delete [] zFull;
    zBase = zFull = 0;
  }
}

/*
** This structure holds state information used for renumbering.
*/
struct RenumberState {
  const char *zPrevName;  /* The name of the previous node */
  int iPrev;              /* index of the previous node */
  int iFirst;             /* index of the first node with a particular name */
  int nStep;              /* step indices by this amount */
  int mxIdx;              /* maximum value of any index */
};

/*
** Renumber all the indices of the children of *this so that they
** are in an evenly spaced accending order.  Return the largest
** index value.
*/
PRIVATE int XmlNode::renumberChildren(int iFirst, int nStep){
  struct RenumberState rns;
  if( pChild==0 ) return 0;
  rns.zPrevName = 0;
  rns.iPrev = 0;
  rns.iFirst = iFirst;
  rns.nStep = nStep;
  rns.mxIdx = 0;
  pChild->renumberSiblings(&rns);
  return rns.mxIdx;
}

/*
** Renumber all siblings of *this according to the specifications
** in *p.
*/
PRIVATE void XmlNode::renumberSiblings(struct RenumberState *p){
  if( pLeft ) pLeft->renumberSiblings(p);
  if( p->zPrevName==0 || strcmp(zName,p->zPrevName)!=0 ){
     p->zPrevName = zName;
     p->iPrev = p->iFirst - p->nStep;
  }
  p->iPrev += p->nStep;
  if( iIndex!=p->iPrev ){
    XmlCursor cur;
    cur = this;
    if( (flags & XML_NODE_TRACE)==0 ){
      flags |= XML_NODE_TRACE;
      int curInvalid = 0;
      if( invokeTrace(&cur, 0, XML_TRACE_INDEX1, &curInvalid) ){
        return;  // This node was removed by a trace routine
      }
      flags &= ~XML_NODE_TRACE;
    }
    iIndex = p->iPrev;
    eraseCachedNames();
    if( (flags & XML_NODE_TRACE)==0 ){
      flags |= XML_NODE_TRACE;
      int curInvalid = 0;
      if( invokeTrace(&cur, 0, XML_TRACE_INDEX2, &curInvalid) ){
        return;  // This node was removed by a trace routine
      }
      flags &= ~XML_NODE_TRACE;
    }
  }
  if( iIndex>p->mxIdx ){
    p->mxIdx = iIndex;
  }
  if( pRight ) pRight->renumberSiblings(p);
  if( pChild ){
    int mx = renumberChildren(p->iFirst, p->nStep);
    if( mx>p->mxIdx ){
      p->mxIdx = mx;
    }
  }
}

/*
** Change the index numbers on all descendents of pOther so that
** they match the index numbers on descendents of *this.
*/
PRIVATE void XmlNode::alignIndices(XmlNode *pOther){
  XmlNode *pMyChild = 0;
  XmlNode *pOtherChild = 0;
  XmlNode *pNextChild = 0;
  const char *zPrevName = 0;
  const char *zNextName = 0;
  int iPrev = 0;
  XmlCursor cur;

  if( pChild==0 || pOther->pChild==0 ) return;
  pMyChild = pChild->leftmostSibling();
  pOtherChild = pOther->pChild->leftmostSibling();
  while( pMyChild && pOtherChild ){
    pNextChild = pOtherChild->next();
    zNextName = pOtherChild->zName;

    int c = strcmp(pMyChild->zName, pOtherChild->zName);
    if( c==0 ){
      pMyChild->alignIndices(pOtherChild);

      if( pOtherChild->iIndex!=pMyChild->iIndex ){
        cur = pOtherChild;
        if( (pOtherChild->flags & XML_NODE_TRACE)==0 ){
          pOtherChild->flags |= XML_NODE_TRACE;
          int curInvalid = 0;
          if( pOtherChild->invokeTrace(&cur, 0, XML_TRACE_INDEX1, 
              &curInvalid) ){
            pOtherChild = 0;  // Trace deleted pOtherChild node
            goto case1;
          }
          pOtherChild->flags &= ~XML_NODE_TRACE;
        }
        iPrev = pOtherChild->iIndex = pMyChild->iIndex;
        pOtherChild->eraseCachedNames();
        cur = pOtherChild;
        if( (pOtherChild->flags & XML_NODE_TRACE)==0 ){
          pOtherChild->flags |= XML_NODE_TRACE;
          int curInvalid = 0;
          if( pOtherChild->invokeTrace(&cur, 0, XML_TRACE_INDEX2, 
              &curInvalid) ){
            pOtherChild = 0;  // Trace deleted pOtherChild node
            goto case1;
          }
          pOtherChild->flags &= ~XML_NODE_TRACE;
        }
      } else {
        iPrev = pOtherChild->iIndex;
      }

case1:
      zPrevName = zNextName;
      pMyChild = pMyChild->next();
      pOtherChild = pNextChild;
    }else if( c<0 ){
      pMyChild = pMyChild->next();
      zPrevName = 0;
    }else{
      if( zPrevName && strcmp(pOtherChild->zName, zPrevName)==0 ){
        if( pOtherChild->iIndex>iPrev ){
          zPrevName = 0;
        }else{
          if( ++iPrev!=pOtherChild->iIndex ){
            cur = pOtherChild;
            if( (pOtherChild->flags & XML_NODE_TRACE)==0 ){
              pOtherChild->flags |= XML_NODE_TRACE;
              int curInvalid = 0;
              if( pOtherChild->invokeTrace(&cur, 0, XML_TRACE_INDEX1, 
                  &curInvalid) ){
                pOtherChild = 0;  // Trace deleted pOtherChild node
                goto case2;
              }
              pOtherChild->flags &= ~XML_NODE_TRACE;
              goto case2;
            }
            pOtherChild->iIndex = iPrev;
            pOtherChild->eraseCachedNames();
            cur = pOtherChild;
            if( (pOtherChild->flags & XML_NODE_TRACE)==0 ){
              pOtherChild->flags |= XML_NODE_TRACE;
              int curInvalid = 0;
              if( pOtherChild->invokeTrace(&cur, 0, XML_TRACE_INDEX2, 
                  &curInvalid) ){
                pOtherChild = 0;  // Trace deleted pOtherChild node
                goto case2;
              }
              pOtherChild->flags &= ~XML_NODE_TRACE;
            }
          }
        }
      }
case2:
      pOtherChild = pNextChild;
    }
  }
  pNextChild = 0;
  while( pOtherChild && zPrevName && strcmp(pOtherChild->zName, zPrevName)==0 ){
    pNextChild = pOtherChild->next();

    if( pOtherChild->iIndex>iPrev ){
      zPrevName = 0;
    }else{
      if( ++iPrev!=pOtherChild->iIndex ){
        cur = pOtherChild;
        if( (pOtherChild->flags & XML_NODE_TRACE)==0 ){
          pOtherChild->flags |= XML_NODE_TRACE;
          int curInvalid = 0;
          if( pOtherChild->invokeTrace(&cur, 0, XML_TRACE_INDEX1, 
              &curInvalid) ){
            pOtherChild = 0;  // Trace deleted pOtherChild node
            goto case3;
          }
          pOtherChild->flags &= ~XML_NODE_TRACE;
        }
        pOtherChild->iIndex = iPrev;
        pOtherChild->eraseCachedNames();
        cur = pOtherChild;
        if( (pOtherChild->flags & XML_NODE_TRACE)==0 ){
          pOtherChild->flags |= XML_NODE_TRACE;
          int curInvalid = 0;
          if( pOtherChild->invokeTrace(&cur, 0, XML_TRACE_INDEX2, 
              &curInvalid) ){
            pOtherChild = 0;  // Trace deleted pOtherChild node
            goto case3;
          }
          pOtherChild->flags &= ~XML_NODE_TRACE;
        }
      }
    }
case3:
    pOtherChild = pNextChild;
  }
}

PUBLIC XmlCursor::XmlCursor(){
  pNode = &xmlRoot;
  moveOk = 1;
  xmlRoot.ref();
  pAttr = 0;
  attrMoveOk = 0;
}
PUBLIC XmlCursor::XmlCursor(const XmlCursor &other){
  pNode = other.pNode;
  moveOk = other.moveOk;
  pNode->ref();
  pAttr = other.pAttr;
  attrMoveOk = other.attrMoveOk;
}
PUBLIC XmlCursor::XmlCursor(XmlNode *pOtherNode){
  pNode = pOtherNode;
  moveOk = 1;
  pNode->ref();
  pAttr = 0;
  attrMoveOk = 0;
}
PUBLIC XmlCursor::XmlCursor(const char *zPath){
  pNode = &xmlRoot;
  pNode->ref();
  moveOk = 1;
  move(zPath);
  pAttr = 0;
  attrMoveOk = 0;
}
PUBLIC XmlCursor::~XmlCursor(){
  pNode->unref();
  pNode = 0;
  pAttr = 0;
}

/*
** The assignment operator when a cursor is on both the left-hand
** and right-hand side.
**
** In an assignment, only the node pointed to is copied.  The
** XmlCursor.moveOk flag is NOT copied.
*/
PUBLIC XmlCursor &XmlCursor::operator=(const XmlCursor &from){
  from.pNode->ref();
  pNode->unref();
  pNode = from.pNode;
  // moveOk = from.moveOk;
  pAttr = from.pAttr;
  // attrMoveOk = from.attrMoveOk;
  return *this;
}

/*
** The assignment operator when a cursor is on the left-hand side
** and a node is on right-hand side.
*/
PUBLIC XmlCursor &XmlCursor::operator=(XmlNode *node){
  node->ref();
  pNode->unref();
  pNode = node;
  moveOk = 1;
  pAttr = 0;
  attrMoveOk = 0;
  return *this;
}

/*
** Return true if a node with the given pathname exists.
*/
PUBLIC int XmlCursor::exists(const char *zPath) const {
  XmlNode *pOther = pNode->findByPath(zPath);
  return pOther!=0;
}

/*
** Return zero if the subtree pointed to by this cursor is identical
** than that pointed to by pOther, nonzero otherwise.
*/
static int DeepCompare(const XmlCursor*, const XmlCursor*);
PUBLIC int XmlCursor::Compare(const XmlCursor* pOther) const {
  return DeepCompare(this, pOther);
}

/*
** Compare two subtrees.  Return true if they differ in any way.
** return false if they are identical.
*/
static int DeepCompare(const XmlCursor *pLeft, const XmlCursor *pRight){
  XmlCursor left = *pLeft;
  XmlCursor right = *pRight;

  for(left.first_attr(); left.attr_ok(); left.next_attr()){
    if( !right.attr_exists(left.getAttrName()) ){return 1;}
    if( strcmp(right.getAttribute(left.getAttrName()), 
               left.getAttribute()) ){return 1;}
  }
  for(right.first_attr(); right.attr_ok(); right.next_attr()){
    if( !left.attr_exists(right.getAttrName()) ){return 1;}
    if( strcmp(left.getAttribute(right.getAttrName()), 
               right.getAttribute()) ){return 1;}
  }

  left = *pLeft;
  right = *pRight;
  left.move_first();
  right.move_first();
  while( left.ok() && right.ok() ){
    if( strcmp(left.basename(), right.basename()) ) return 1;
    if( left.is_leaf() ){
      if( !right.is_leaf() ) return 1;
      if( !left.equal(left, right) ) return 1;
    }else if( right.is_leaf() ){
      return 1;
    }else if( DeepCompare(&left, &right) ){
      return 1;
    }
    for(left.first_attr(); left.attr_ok(); left.next_attr()){
      if( !right.attr_exists(left.getAttrName()) ){return 1;}
      if( strcmp(right.getAttribute(left.getAttrName()), 
                 left.getAttribute()) ){return 1;}
    }
    for(right.first_attr(); right.attr_ok(); right.next_attr()){
      if( !left.attr_exists(right.getAttrName()) ){return 1;}
      if( strcmp(left.getAttribute(right.getAttrName()), 
                 right.getAttribute()) ){return 1;}
    }
    left.move_next();
    right.move_next();
  }
  return left.ok() || right.ok();
}

/*
** Move to the named node and return TRUE, if such a node exists.  
** If the node does not exist, return FALSE and do no move the cursor.
*/
PUBLIC int XmlCursor::move(const char *zPath){
  XmlNode *pOther = pNode->findByPath(zPath);
  if( pOther==0 ){
    moveOk = 0;
    return 0;
  }
  pOther->ref();
  pNode->unref();
  pNode = pOther;
  moveOk = 1;
  attrMoveOk = 0;
  return 1;
}

/*
** Move the cursor up to the parent node and return TRUE.  If this
** node is the root of the XML tree (and thus cannot be moved up)
** then the cursor is unchanged and this routine returns FALSE.
*/
PUBLIC int XmlCursor::move_up(){
  int rc;
  if( pNode->pParent ){
    XmlNode *pOld = pNode;
    pNode = pNode->pParent;
    pNode->ref();
    pOld->unref();
    pOld = 0;
    rc = 1;
    attrMoveOk = 0;
  }else{
    rc = 0;
  }
  moveOk = rc;
  return rc;
}

/*
** Move the cursor down to first child node and return TRUE.  If this
** node has no children (and thus cannot be moved down)
** then the cursor is unchanged and this routine returns FALSE.
*/
PUBLIC int XmlCursor::move_first(){
  int rc;
  if( pNode->pChild ){
    XmlNode *pOld = pNode;
    pNode = pNode->pChild->leftmostSibling();
    pNode->ref();
    pOld->unref();
    pOld = 0;
    rc = 1;
    attrMoveOk = 0;
  }else{
    rc = 0;
  }
  moveOk = rc;
  return rc;
}

/*
** Move the cursor horizontally to the next sibling and return TRUE.
** If this node is the last sibling and thus cannot be moved further
** to the right, then the cursor is unchanged and this routine
** returns FALSE.
*/
PUBLIC int XmlCursor::move_next(){
  int rc;
  XmlNode *pNew = pNode->next();
  if( pNew ){
    XmlNode *pOld = pNode;
    pNode = pNew;
    pNode->ref();
    pOld->unref();
    pOld = 0;
    rc = 1;
    attrMoveOk = 0;
  }else{
    rc = 0;
  }
  moveOk = rc;
  return rc;
}

/*
** Return true iff the node we're pointing to is ancestral to the node
** pOther points to.
*/
PUBLIC int XmlCursor::is_ancestor(const XmlCursor* pOther) const {
  return pNode->is_ancestor(pOther->pNode);
}

/*
** Move the cursor to the next attribute and return TRUE.
** If we are at the last attribute, don't move the cursor
** and return FALSE.
*/
PUBLIC int XmlCursor::next_attr(){
  int rc;
  if( pAttr==0 ){
    attrMoveOk = 0;
    return 0;
  }
  XmlAttribute *pNew = pAttr->pNext;
  if( pNew ){
    pAttr = pNew;
    rc = 1;
  }else{
    rc = 0;
  }
  attrMoveOk = rc;
  return rc;
}

/*
** Move the cursor to the first attribute and return TRUE.
** If this node has no attributes, the cursor is unchanged
** and this routine returns FALSE.
*/
PUBLIC int XmlCursor::first_attr(){
  int rc;
  if( pNode->pAttrList ){
    pAttr = pNode->pAttrList;
    rc = 1;
  }else{
    rc = 0;
  }
  attrMoveOk = rc;
  return rc;
}

/*
** Create a new node in the XML tree.  The name of the new node
** is specified by the argument.  Intermediate nodes are also
** created as necessary. 
*/
PRIVATE XmlNode *XmlCursor::createNode(const char *zPath){
  assert( zPath!=NULL );
  int i;
  XmlNode *pRent = pNode;
  const char *zBaseName;

  char *zTrimPath = new char[strlen(zPath)+1];
  strcpy(zTrimPath, zPath);
  for(i=strlen(zTrimPath)-1; i!=0 && zTrimPath[i]=='/'; i--){
    zTrimPath[i] = 0;
  }
  // zPath = zTrimPath;  ... don't do this kind of thing

  for(i=0; zTrimPath[i] && zTrimPath[i]!='/'; i++){}
  if( zTrimPath[i]==0 ){
    zBaseName = zTrimPath;
  }else{
    for(i=strlen(zTrimPath)-1; zTrimPath[i]!='/'; i--){}
    zBaseName = &zTrimPath[i+1];
    while( i>0 && zTrimPath[i]=='/' ){ i--; }
    char *zParentPath = new char[i+2];
    sprintf(zParentPath, "%.*s", i+1, zTrimPath);
    pRent = pNode->findByPath(zParentPath);
    if( pRent==0 ){
      pRent = createNode(zParentPath);  // May fail due to trace proc
      if( pRent && pRent->unref() ){
        pRent = 0;
      }
    }
    delete [] zParentPath; zParentPath = 0;
  }

  /* If creating the parent failed (due to trace proc), don't try to
  ** create this node.
  */
  if( !pRent ) return 0;
  XmlNode *pNewNode = new XmlNode(zBaseName, -1);
  assert( pNewNode!=0 );
  pNewNode->ref();
  if( pNewNode->insert(pRent) ){  // Insert failed, due to trace proc
    pNewNode = 0;
  }
  delete [] zTrimPath; zTrimPath = 0;
  return pNewNode;
}

/*
** Create a new entry in the XML tree and return the name of the new
** entry.  If the name given by zPath contains explicit indices
** (for example /xyz:5/abc:11) then an effort is made to create
** entries with those same indices.  But if a collision occurs (for
** example if /xyz:5 already exists), then the indices are advanced to
** the next unused index.
**
** The cursor is left pointing at the newly created node.  To create a
** node without moving the cursor, see the mkdir() method.
**
** The string returned is the full pathname of the newly created entry.
**
** This proc may return 0 if the creation failed.  A creation can fail
** because of the invocation of a trace routine.
*/
PUBLIC const char *XmlCursor::create(const char *zPath){
  assert( zPath!=NULL );
  XmlNode *pNewNode = createNode(zPath);
  if( !pNewNode ) return 0;   // Trace proc may have killed the new node
  pNode->unref();
  pNode = pNewNode;
  const char* zFull = pNode->fullname();
  moveOk = (zFull && zFull[0]);
  return zFull;
}

/*
** Reset the index auto-generator to 0.
**
** We'd like to just pass a default argument to insert() rather than using
** this reset_counter hack, but that confuses makeheaders.
*/
PUBLIC void XmlCursor::resetCounter(){
  pNode->reset_counter=1;
  pNode->insert(NULL);    // will not fail. this is a hack.
}

/*
** This version of addAttribute is used when we set the value of the
** attribute currently being visited. If we're not visiting an attribute,
** return FALSE.
*/
PUBLIC int XmlCursor::addAttribute(const char *zValue){
  if( !zValue || !attr_ok() ){ return 0; }
  pNode->changeAttribute(pAttr, zValue);
  return 1;
}

/*
** Add a new attribute (or modify an existing one if they have the same name)
** to the current node.
*/
PUBLIC int XmlCursor::addAttribute(const char *zName, const char *zValue){
  XmlAttribute *pTemp;

  if( !zValue ){ return 0; }
  if( !zName ){ return addAttribute(zValue); }
  pTemp = pNode->pAttrList;
  if( !pTemp ){ pNode->createAttribute(zName, zValue); }
  while( pTemp ){
    if( !strcmp(pTemp->zName, zName) ){
      pNode->changeAttribute(pTemp, zValue); 
      break;
    }
    if( !pTemp->pNext ){
      pNode->createAttribute(zName, zValue);
      break;
    }
    pTemp = pTemp->pNext;
  }
  return 1;
}

/*
** Overloaded to help with adding attributes when parsing XML
*/
PUBLIC int XmlCursor::addAttribute(
  const char *zName,                // Pointer to the name of the attribute
  int nName,                        // Length of the name
  const char *zValue,               // Pointer to the value of the attribute
  int nValue                        // Length of the value
){
  int r;
  char *zNewName = new char[nName+1];
  char *zNewValue = new char[nValue+1];
  strncpy(zNewName, zName, nName);
  zNewName[nName] = 0;
  strncpy(zNewValue, zValue, nValue);
  zNewValue[nValue] = 0;
  XmlNode::decodeXmlString(zNewValue);
  r = addAttribute(zNewName, zNewValue);
  delete [] zNewName; zNewName = 0;
  delete [] zNewValue; zNewValue = 0;
  return r;
}

/*
** Remove an attribute from the attribute list. If the attribute does
** not exist, do nothing.
*/
PUBLIC int XmlCursor::removeAttribute(const char *zName){
  XmlAttribute *pPrev;
  XmlAttribute *pTemp;
  
  if( !pNode->pAttrList ){ return 0; }
  if( !strcmp(pNode->pAttrList->zName, zName) ){
    pTemp = pNode->pAttrList;
    pNode->pAttrList = pNode->pAttrList->pNext;
    pTemp->pNext = 0;
    delete pTemp; pTemp = 0;
    if( (pNode->flags & XML_NODE_TRACE)==0 ){
      pNode->flags |= XML_NODE_TRACE;
      int curInvalid = 0;
      if( pNode->invokeTrace(this, 0, XML_TRACE_MODIFY, &curInvalid) ){
        pNode = 0;    // Trace deleted pNode
      }else{
        pNode->flags &= ~XML_NODE_TRACE;
      }
    }
    return 1;
  }
  pPrev = pNode->pAttrList;
  pTemp = pNode->pAttrList->pNext;
  while( pTemp ){
    if( !strcmp(pTemp->zName, zName) ){
      pPrev->pNext = pTemp->pNext;
      pTemp->pNext = 0;
      delete pTemp; pTemp = 0;
      if( (pNode->flags & XML_NODE_TRACE)==0 ){
        pNode->flags |= XML_NODE_TRACE;
        int curInvalid = 0;
        if( pNode->invokeTrace(this, 0, XML_TRACE_MODIFY, &curInvalid) ){
          pNode = 0;    // Trace deleted pNode
        }else{
          pNode->flags &= ~XML_NODE_TRACE;
        }
      }
      return 1;
    }
    pPrev = pTemp;
    pTemp = pTemp->pNext;
  }
  return 0;
}

/*
** Remove the attribute currently being visited.
*/
PUBLIC int XmlCursor::removeAttribute(void){
  if( pAttr && attr_ok() ) return removeAttribute(pAttr->zName);
  return 0;
}

/*
** Return the value of the attribute given by *zName or the value of the
** attribute currently being visited if *zName is null.
*/
PUBLIC const char *XmlCursor::getAttribute(const char *zName) const {
  if( !zName ){
    if( attr_ok() ) return pAttr->zValue;
    return 0;
  }
  XmlAttribute *pTemp = pNode->pAttrList;
  while( pTemp ){
    if( strcmp(pTemp->zName, zName)==0 ){ return pTemp->zValue; }
    pTemp = pTemp->pNext;
  }
  return 0;
}
PUBLIC const char *XmlCursor::getAttribute(void) const {
  return getAttribute(0);
}

/*
** Return the name of the attribute currently being visited or an empty string
** if no attribute is being visited.
*/
PUBLIC const char *XmlCursor::getAttrName(){
  if( pAttr && attr_ok() ) return pAttr->zName;
  return "";
}

/*
** Return TRUE if the attribute zName exists.
*/
PUBLIC int XmlCursor::attr_exists(const char *zName){
  if( !zName || !*zName ) return 0;
  XmlAttribute *pTemp = pNode->pAttrList;
  while( pTemp ){
    if( !strcmp(pTemp->zName, zName) ) return 1;
    pTemp = pTemp->pNext;
  }
  return 0;
}

/*
** Move the cursor to the node named in the argument.  Create the
** node if it does not already exist.
**
** The creation may fail due to a trace proc invocation.
*/
PUBLIC const char *XmlCursor::moveOrCreate(const char *zPath){
  move(zPath);
  if( !ok() ){
    if( !create(zPath) ) return 0;
  }
  return fullname();
}

/*
** Destroy the node named by the zPath pathname.  Also delete
** all of the children of that node.
**
** This routine will not delete the root node "/".  If you try
** to delete the root node, all the children of the root node
** get deleted but the root node itself is not deleted.
*/
PUBLIC int XmlCursor::unlink(const char *zPath){
  XmlNode *pOld = pNode->findByPath(zPath);
  if( pOld ){
    if( pOld->pParent ){
      pOld->unlink();
      pOld = 0;
    }else{
      // pOld->removeAllTraces();
      if( pOld->pChild ){
        pOld->pChild->unlinkAllSiblings();
        pOld->pChild = 0;
      }
    }
  }
  return 0;
}

/*
** Destroy the node at which the cursor is pointing.  Leave the cursor
** pointing at the parent node.
*/
PUBLIC int XmlCursor::unlink(){
  if( !pNode->pParent ) return 0;   // Never delete the root node
  XmlNode *pOld = pNode;
  move_up();
  pOld->unlink();
  pOld = 0;
  return 0;
}

/*
** Return information on the memory used by the node the cursor
** is pointing to all all decendent nodes.
*/
PUBLIC void XmlCursor::memUsed(int *aMem){
  pNode->memUsed(aMem, 0);
}

/*
** Renumber all nodes that are children of the node that the
** cursor is point to.  Return the largest index in the renumbering.
*/
PUBLIC int XmlCursor::renumber(int iFirst, int nStep){
  return pNode->renumberChildren(iFirst, nStep);
}

/*
** Align the index numbers for all children of the pOther cursor with
** the indices on the children of this cursor.  The pOther cursor is
** the one that changes.
*/
PUBLIC void XmlCursor::align(XmlCursor *pOther){
  pNode->alignIndices(pOther->pNode);
}

#if LOCAL_INTERFACE
/*
** This class is used to create a string that grows without
** bound.  As new text is added, additional memory is allocated.
** The memory allocation doubles each time so that the total
** number of allocations is the log() of the string length.
*/
class XmlDynString {
private:
  int nAlloc;    // Number of characters allocated in zBuf[] 
  int nUsed;     // Number of slots of zBuf[] actually used
  char *zBuf;    // The dynamic string
public:
};
#endif

PUBLIC XmlDynString::XmlDynString(){ nAlloc = nUsed = 0; zBuf = 0; };
PUBLIC XmlDynString::~XmlDynString(){ if( zBuf ){ ckfree(zBuf); }};

/*
** Make the dynamic string big enough to hold N bytes.
** return 0 on success and 1 if there is a failure to
** allocate memory.
*/
PRIVATE int XmlDynString::resize(int N){
  if( N + 1 >= nAlloc ){
    if( nAlloc==0 ){
      nAlloc = N + 100;
      zBuf = ckalloc( nAlloc );
    }else{
      nAlloc += nAlloc + N;
      zBuf = ckrealloc(zBuf, nAlloc );
    }
    if( zBuf==0 ){
      nAlloc = nUsed = 0;
      return 1;
    }
  }
  return 0;
}

/*
** Append text to the string
*/
PUBLIC int XmlDynString::append(const char *zText, int nText){
  if( zText==0 ) return nUsed;
  if( nText<0 ) nText = strlen(zText);
  if( resize(nUsed + nText) ) return 0;
  memcpy(&zBuf[nUsed], zText, nText);
  nUsed += nText;
  return nUsed;
}

/*
** Append text to the string that is encoded for XML.  Most characters
** are copied straight through without change.  But the following
** characters are changed:
**
**     FROM            TO
**     -----          ------
**      <              &lt;
**      >              &gt;
**      &              &amp;
**      '              &apos;
**      "              &quot;
*/
PUBLIC int XmlDynString::appendXml(const char *zText, int nText){
  int i;
  if( nText<0 ) nText = strlen(zText);
  while( nText>0 ){
    int c = 0;
    for(
        i=0; 
        i<nText && (c=zText[i])!='<' && c!='&' && c!='>' && c!='"'
        && c!='\'' && isprint(c);
        i++
    ){}
    if( i>0 ) append(zText, i);
    if( i>=nText ) break;
    switch( c ){
      case '<':  append("&lt;",4);   break;
      case '>':  append("&gt;",4);   break;
      case '&':  append("&amp;",5);  break;
      case '\'': append("&apos;",6); break;
      case '"':  append("&quot;",6); break;
      default:   
        /* The character's not printable.  Skip it. */
      break;
    }
    zText += i+1;
    nText -= i+1;
  }
  return nUsed;
}

/*
** Append spaces to a dynamic string.
*/
PUBLIC int XmlDynString::appendSpace(int amt){
  if( resize(nUsed + amt) ) return 0;
  memset(&zBuf[nUsed], ' ', amt);
  nUsed += amt;
  return nUsed;
}

/*
** Truncate the string and return a pointer to the text.
** the calling function should delete this text by passing
** it to ckfree().
*/
PUBLIC char *XmlDynString::finalize(){
  char *zResult;
  if( nAlloc==0 ){
    return 0;
  }
  if( nAlloc > nUsed+1 ){
    nAlloc = nUsed+1;
    zBuf = ckrealloc(zBuf, nAlloc);
  }
  zResult = zBuf;
  zResult[nUsed] = 0;
  zBuf = 0;
  nAlloc = nUsed = 0;
  return zResult;
}

/*
** Render part of the XML tree into an XmlDynString.  Indent the
** generated text by the amount given.
**
** This is a helper routine.  See below for the real render()
** method.
*/
PRIVATE void XmlCursor::render(
  class XmlDynString *pStr,    // Write the XML into this string
  int indent                   // Indent by this many levels
){
  pStr->appendSpace(indent);
  pStr->append("<", 1);
  pStr->append(name(),-1);
  XmlAttribute *pTemp = pNode->pAttrList;
  while( pTemp ){
    pStr->append(" ", 1);
    pStr->append(pTemp->zName, -1);
    pStr->append("=\"", 2);
    pStr->appendXml(pTemp->zValue, -1);
    pStr->append("\"", 1);
    pTemp = pTemp->pNext;
  }
  pStr->append(">", 1);
  if( is_leaf() ){
    pStr->appendXml(get(), length());
  }else{
    pStr->append("\n",1);
    XmlCursor cur2 = *this;
    for(cur2.move_first(); cur2.ok(); cur2.move_next()){
      cur2.render(pStr, indent+2);
    }
    pStr->appendSpace(indent);
  }
  pStr->append("</", 2);
  pStr->append(name(), -1);
  pStr->append(">\n", 2);
}

/*
** Generate XML text into memory obtained from ckalloc().  The
** calling function is responsible for freeing the memory when
** done.  If bStrip is true, strip off the top-level tag.
*/
PUBLIC char *XmlCursor::render(int indent, int bStrip){
  if( !ok() ){ return NULL; }
  XmlDynString sStr;
  if( pNode->pParent && !bStrip ){
    render(&sStr, indent);
  }else{
    XmlCursor cur2 = *this;
    for(cur2.move_first(); cur2.ok(); cur2.move_next()){
      cur2.render(&sStr, indent);
    }
  }
  return sStr.finalize();
}
PUBLIC char *XmlCursor::render(int indent){
  return render(indent, 0);
}

/* An array to map all upper-case characters into their corresponding
** lower-case character. 
*/
static unsigned char UpperToLower[] = {
      0,  1,  2,  3,  4,  5,  6,  7,  8,  9, 10, 11, 12, 13, 14, 15, 16, 17,
     18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35,
     36, 37, 38, 39, 40, 41, 42, 43, 44, 45, 46, 47, 48, 49, 50, 51, 52, 53,
     54, 55, 56, 57, 58, 59, 60, 61, 62, 63, 64, 97, 98, 99,100,101,102,103,
    104,105,106,107,108,109,110,111,112,113,114,115,116,117,118,119,120,121,
    122, 91, 92, 93, 94, 95, 96, 97, 98, 99,100,101,102,103,104,105,106,107,
    108,109,110,111,112,113,114,115,116,117,118,119,120,121,122,123,124,125,
    126,127,128,129,130,131,132,133,134,135,136,137,138,139,140,141,142,143,
    144,145,146,147,148,149,150,151,152,153,154,155,156,157,158,159,160,161,
    162,163,164,165,166,167,168,169,170,171,172,173,174,175,176,177,178,179,
    180,181,182,183,184,185,186,187,188,189,190,191,192,193,194,195,196,197,
    198,199,200,201,202,203,204,205,206,207,208,209,210,211,212,213,214,215,
    216,217,218,219,220,221,222,223,224,225,226,227,228,229,230,231,232,233,
    234,235,236,237,238,239,240,241,242,243,244,245,246,247,248,249,250,251,
    252,253,254,255
};


/*
** Some systems have strincmp().  Others have strcasencmp().  Because
** there is no consistency, we will define our own.
*/
static int StrNICmp(const char *zLeft, const char *zRight, int N){
  register unsigned char *a, *b;
  a = (unsigned char *)zLeft;
  b = (unsigned char *)zRight;
  while( N-- > 0 && *a!=0 && UpperToLower[*a]==UpperToLower[*b]){ a++; b++; }
  return N<0 ? 0 : *a - *b;
}

/*
** Token types found by the XML tokenizer
*/
#define TOKEN_SPACE   0     // Space between markup
#define TOKEN_TEXT    1     // CDATA, including leading and trailing space
#define TOKEN_MARKUP  2     // XML elements
#define TOKEN_UNKNOWN 3     // End of file, usually
#define TOKEN_ERROR   4     // Some kind of error

/*
** Find the next token of XML.  Return the length of the token
** and write the token type into *pType.
*/
static int getXmlToken(const char *zText, int nText, int *pType){
  int i = 1;
  if( zText==0 || nText==0 ){
    *pType = TOKEN_UNKNOWN;
    return 0;
  }
  if( zText[0]=='<' ){
    *pType = TOKEN_MARKUP;
    while( i<nText && zText[i] && zText[i]!='>' ){ i++; }
    if( i<nText && zText[i]=='>' ){ i++; }
    else{ *pType = TOKEN_ERROR; }
    return i;
  }
  if( isspace(zText[0]) ){
    while( i<nText && isspace(zText[i]) ){ i++; }
    if( i==nText || zText[i]=='<' || zText[i]==0 ){
      *pType = TOKEN_SPACE;
      return i;
    }
  }
  *pType = TOKEN_TEXT;
  while( i<nText && zText[i] && zText[i]!='<' ){ i++; }
  /* while( i>0 && isspace(zText[i-1]) ){ i--; } */
  return i;
}

/*
** Parse the given XML text into the tree at the position of the
** cursor.  In other words, all new XML elements become children
** of the current node in the tree.
**
** If the XML is well-formed, the cursor should be left pointing at
** the same node it started on.  If the XML is not well-formed, the
** cursor might be left pointing most anywhere.  Ill-formed XML is 
** parsed as best as we can, but there are no guarantees about what
** will happen.  The parser will not core, but it might fill the
** XML tree with some strange content.
**
** If zOnly is not NULL, then only the first toplevel element
** whose name matches zOnly will be parsed.
*/
PUBLIC int XmlCursor::parse(const char *zText, int nText, const char *zOnly){
  int level = 0;
  XmlNode *pTemp = 0;
  if( nText==0 ){
    pNode->set("", 0);
    return 0;
  }
  if( nText==-1 ) nText = strlen(zText);
  while( nText>0 ){
    int tokenType;
    int tokenLen = getXmlToken(zText, nText, &tokenType);
    switch( tokenType ){
      case TOKEN_SPACE: {
        /* White space is ignored in this application */
        break;
      }
      case TOKEN_TEXT: {
        pNode->set(zText, tokenLen);
        pNode->decodeXml();
        break;
      }
      case TOKEN_MARKUP: {
        if( tokenLen<3 ) break;
        if( zText[1]=='/' && isalnum(zText[2]) ){
          if( level<=0 ) break;
          move_up();
          level--;
          if( level==0 ){
            if( pTemp!=0 ){
              pTemp->unlink();
              pTemp = 0;
            }else if( zOnly!=0 ){
              nText = 0;
            }
          }
          break;
        }
        if( !isalnum(zText[1]) ){
          break;
        }
        level++;
        int i;
        for(i=1; i<tokenLen && (isalnum(zText[i]) || zText[i]=='_' ||
              zText[i]=='-' || zText[i]==':'); i++){}
        XmlNode *pNew = new XmlNode(&zText[1], i-1);
        pNew->ref();
        if( pNew->insert(pNode) ){  // Insertion failed due to trace
          pNew = 0;
          return 0;
        }

        XmlNode* pTemp2 = pNode;
        pNode = pNew;
        /* Check for attributes. NOTE: Improperly formatted XML
        ** could yield unexpected results.
        */
        int j,k;
        while( i<tokenLen ){
          while( i<tokenLen && isspace(zText[i]) ){ i++; }
          j = i;
          while( j<tokenLen && zText[j]!='=' ){ j++; }
          k = j+2;
          while( k<tokenLen && zText[k]!='\"' ){ k++; }
          if( k<tokenLen ){
            addAttribute(&zText[i], j-i, &zText[j+2], k-j-2);
          }
          i = k+1;
        }
        pNode = pTemp2;
        if( tokenLen<4 || zText[tokenLen-2]!='/' ){
          pNode->unref();
          pNode = pNew;
        }else{
          if( pNew->unref() ){
            pNew = 0;
          }
        }

        assert( pNew!=0 );
        if( level==1 && zOnly && strcmp(zOnly,pNew->name())!=0 ){
          pTemp = pNew;
        }
        break;
      }
      default: {
        break;
      }
    }
    if( tokenLen==0 ) break;
    nText -= tokenLen;
    zText += tokenLen;
  }
  return 0;
}

/*
** Test an XML string for well-formedness.  All begin element tags should
** have a corresponding end element tag.
**
** Return zero if there are no errors.  Return positives if errors are
** found.  If *pzErrMsg is not zero, it is made to point to a static
** string that describes the error.  *piOffset will be set to the offset
** of the beginning of the error in zText.  *piLength will hold the 
** length of the error.
*/
STATIC PUBLIC int XmlCursor::check(
  const char *zText,       // XML text to be checked
  int nText,               // Number of bytes of text in zText
  const char **pzErrMsg,   // Description of the error
  int *piOffset,           // Start of error from beginning of zText
  int *piLength            // Length of the error token
){
  int i = 0;
  int n;
  int depth = 0;
# define mxDepth 1000
  const char *zOrigText = zText;
  int nOrigText = nText;
  const char *zDummy;
  int iDummy;
  const char *zStart;
  int iStart[mxDepth];

  if( pzErrMsg==0 ) pzErrMsg = &zDummy;
  if( piOffset==0 ) piOffset = &iDummy;
  if( piLength==0 ) piLength = &iDummy;
  if( nText==0 ) nText = strlen(zText);
  while( nText>0 ){
    int tokenType;
    int tokenLen = getXmlToken(zText, nText, &tokenType);
    switch( tokenType ){
      case TOKEN_TEXT:
      case TOKEN_SPACE: {
        break;
      }
      case TOKEN_MARKUP: {
        if( tokenLen<3 ) break;
        if( zText[1]=='/' ){
          if( !isalnum(zText[2]) ){
            *piOffset = i;
            *piLength = tokenLen;
            *pzErrMsg = "malformed markup";
            return 1;
          }
          if( depth<=0 ){
            *piOffset = i;
            *piLength = tokenLen;
            *pzErrMsg = "no beginning to this element";
            return 1;
          }
          for(n = 0; isalnum(zText[n+2]) || zText[n+2]=='_'; n++){}
          depth--;
          zStart = &zOrigText[iStart[depth]];
          if( StrNICmp(&zStart[1], &zText[2], n) ){
            *piOffset = i;
            *piLength = tokenLen;
            *pzErrMsg = "no beginning to this element";
            return 1;
          }
        }else if( !isalnum(zText[1]) ){
          *piOffset = i;
          *piLength = tokenLen;
          *pzErrMsg = "unrecognized markup";
          return 1;
        }else if( zText[tokenLen-2]=='/' ){
          // A null token.  Ex: "<abc/>"  Do nothing
        }else{
          if( depth>=mxDepth ){
            *piOffset = i;
            *piLength = tokenLen;
            *pzErrMsg = "markup nested too deep";
            return 1;
          }
          iStart[depth] = i;
          depth++;
        }
        break;
      }
      case TOKEN_ERROR: {
        *piOffset = i;
        *piLength = tokenLen;
        *pzErrMsg = "unrecognized token";
        return 1;
      }
      default: {
        break;
      }
    }
    if( tokenLen==0 ) break;
    i += tokenLen;
    nText -= tokenLen;
    zText += tokenLen;
  }
  if( depth!=0 ){
    int i = iStart[depth-1];
    int tokenType;
    int tokenLen = getXmlToken(&zOrigText[i], nOrigText-i, &tokenType);
    *piOffset = i;
    *piLength = tokenLen;
    *pzErrMsg = "unterminated element";
    return 1;
  }
  return 0;
}

/*
** Copy all the content of pFrom into pTo
*/
STATIC PUBLIC void XmlCursor::DeepCopy(XmlCursor *pFrom, XmlCursor *pTo){
  XmlCursor from = *pFrom;
  XmlCursor to = *pTo;
  for(from.first_attr(); from.attr_ok(); from.next_attr()){
    to.addAttribute(from.getAttrName(), from.getAttribute());
  }
  if( from.is_leaf() ){
    if( from.is_real() ){
      to.set(from.get_real());
    }else{
      to.set(from.get(),from.length());
    }
  }else{
    for(from.move_first(); from.ok(); from.move_next()){
      if( to.moveOrCreate(from.basename()) ){  // may fail due to trace proc
        DeepCopy(&from, &to);
        to.move_up();
      }
    }
  }
}

/*
** Make the content of *pTo look just like the content of *pFrom.
** Except, if bNoToplevelAttr is true, don't sync the top-level attributes
** of pTo to those of pFrom.
*/
STATIC PUBLIC void XmlCursor::DeepSync(XmlCursor *pFrom, XmlCursor *pTo){
  DeepSync(pFrom, pTo, 0);
}
STATIC PUBLIC void XmlCursor::DeepSync(
  XmlCursor *pFrom, XmlCursor *pTo, int bNoToplevelAttr
){
  XmlCursor from = *pFrom;
  XmlCursor to = *pTo;
  if( !bNoToplevelAttr ){
    for(to.first_attr(); to.attr_ok(); ){
      if( !pFrom->attr_exists(to.getAttrName()) ){
        XmlCursor toDel = to;
        to.next_attr();
        toDel.removeAttribute();
      }else{
        to.next_attr();
      }
    }
    to = *pTo;
    for(from.first_attr(); from.attr_ok(); from.next_attr()){
      to.addAttribute(from.getAttrName(), from.getAttribute());
    }
  }
  if( from.is_leaf() ){
    if( from.is_real() ){
      to.set(from.get_real());
    }else{
      to.set(from.get(), from.length());
    }
  }else{
    for(to.move_first(); to.ok(); ){
      if( !pFrom->exists(to.basename()) ){
        XmlCursor toDel = to;
        to.move_next();
        toDel.unlink();
      }else{
        to.move_next();
      }
    }
    to = *pTo;
    for(from.move_first(); from.ok(); from.move_next()){
      if( to.moveOrCreate(from.basename()) ){  // may fail due to trace proc
        DeepSync(&from, &to, 0);
        to.move_up();
      }
    }
  }
}

/*
** Write the boolean value v into the node named zPath.
**
** The way a boolean value works is:  if 'v' is TRUE, then
** the node's value is set to 1.  If 'v' is FALSE, then
** the node is deleted.
*/
PUBLIC void XmlCursor::set_boolean(const char *zPath, int v){
  if( !zPath || !*zPath ) return;
  if( v ){
    move(zPath);
    if( !ok() ){
      if( !create(zPath) ) return;  // Create failed due to trace
    }
    set(1.0);
  }else{
    unlink(zPath);
  }
}

