static const char rcsid[] = "@(#) $Header: /home/cvs/softools/code/common/xmlp/src/xmlphash.cpp,v 1.28 2011/06/01 14:53:58 eas Exp $";
/*
** Copyright (c) 2005-2008, 2010-2011 7B Software, Inc.
** All rights reserved.
**
** This file is part of XMLP.
**
** XMLP is free software: you can redistribute it and/or modify
** it under the terms of the GNU Lesser General Public License as published by
** the Free Software Foundation, either version 3 of the License, or
** (at your option) any later version.
**
** XMLP is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Lesser General Public License for more details.
**
** You should have received a copy of the GNU Lesser General Public License
** along with XMLP.  If not, see <http://www.gnu.org/licenses/>.
**
** Author contact information:
**   mark.smith@7bsoftware.com
**   http://www.7bsoftware.com
*/

/*
** This file implements indices into the XML tree.  We call these 
** indices "hashes" because in XMLP the word "index" is used for 
** something entirely different.
*/

#include <stdio.h>
#include <string.h>
#include <tcl.h>
#include <assert.h>
#include <inttypes.h>
#include "xmlphash.h"

#if INTERFACE

#include <boost/unordered_map.hpp>
#include <boost/unordered_set.hpp>

#include <string>

/*
** This class implements an index into the xml tree based on some attribute
** or leaf name.  A hash can either be an ATTRIBUTE HASH or a LEAF HASH,
** as given by iType.
**
** We refer to the "salient value" of a node N given hash H (on name H.NAME) as 
** SalientValue(H, N) = :
** 1. The value of xml attribute H.NAME on node N, if H is an attribute hash;
** 2. The value of N, if:
**    a. N's basename is H.NAME, and 
**    b. N is a leaf node, and
**    c. H is a leaf hash;
** 3. Undefined (or NULL), otherwise.
*/
class XmlHash {
private:

#ifndef XMLHASH_STRICT_CHECKS

  /* Define this to 1 if we want to do strict checking on every hash update.
  ** Note that this will probably slow down the program significantly, since
  ** it forces a crawl from a hash root if a descendent of the root is
  ** altered.
  */
  #define XMLHASH_STRICT_CHECKS 0
  #if XMLHASH_STRICT_CHECKS && (!defined(NDEBUG) || !NDEBUG)
    #define SANITY_CHECK(A) StrictCheck(A)
    #define SANITY_CHECK_HASH(A,B) (A)->StrictCheck(B)
  #else
    #define SANITY_CHECK(A)
    #define SANITY_CHECK_HASH(A,B)
  #endif
#endif

  /* These three variables implement a doubly-linked list of hashes.
  */
  static XmlHash* pFirst;
  XmlHash* pNext;
  XmlHash* pPrev;

  /* The root we're indexing and the attribute or leaf name we care about.
  ** zRoot is not used at the moment, other than to set up (and destroy
  ** at destruction time) the xml trace.
  */
  char* zRoot;
  char* zName;

  /* One of XMLHASH_TYPE_*.  Tells us whether we're an attribute hash or
  ** a leaf hash.
  */
  int iType;

  /* True if this is a unique-hash.
  */
  short bUnique;

  /* A set of XML paths.
  */
  typedef boost::unordered_set<std::string> path_set_type;

  /* Map a salient value to a set of paths with that salient value.
  */
  typedef boost::unordered_map<std::string, path_set_type> map_type;

  /* Map an XML path to its salient value.
  */
  typedef boost::unordered_map<std::string, std::string> path2value_type;

  map_type m_map;
  path2value_type m_reverse_map;

  /* The hash context "owner" of this XmlHash.  Used so that we
  ** can invalidate its pointer to this object, when this object
  ** is deleted.
  **
  ** If this object is not "owned" by a HashProcCtx, then the
  ** value of this pointer will be 0.
  */
  HashProcCtx* pHashCtx;

public:

  /* The tcl interpreter we're registered with.
  */
  Tcl_Interp* pInterp;

  /*
  ** Dump a list of all registered hashes.
  */
  static int dump_list(Tcl_Interp* pInterp) {
    fprintf(stderr, "list of hashes:\n");
    Tcl_Obj* pList = Tcl_NewListObj(0, NULL);
    for(XmlHash* p=pFirst; p; p=p->pNext){
      char zName[100];
      snprintf(zName, sizeof zName, "-hash-%p-", p);
      Tcl_ListObjAppendElement(pInterp, pList, Tcl_NewStringObj(zName, -1));
      fprintf(stderr, "  %s\n", zName);
    }
    Tcl_SetObjResult(pInterp, pList);
    return TCL_OK;
  }

  /*
  ** Run a strict check on all hashes.
  */
  static void StrictCheckAll(int chngType){
#if XMLHASH_STRICT_CHECKS && (!defined(NDEBUG) || !NDEBUG)
    for( XmlHash* p=pFirst; p; p=p->pNext ){
      SANITY_CHECK_HASH(p, chngType);
    }
#endif
  }
};

#define XMLHASH_TYPE_LEAF       1
#define XMLHASH_TYPE_ATTR       2

#endif /* INTERFACE */

/* Initialize the static member datum.
*/
XmlHash* XmlHash::pFirst = 0;

/*
** This function is called whenever a change occurs to the xml that will
** interest a hash map.
**
** The caller must ensure that pUserData is valid; that is, it does not
** reference a deleted object.  XmlHash::~XmlHash() is responsible for
** making sure that this function isn't called.
*/
static int hashTraceCallback(
  XmlCursor *pCursor,
  XmlCursor *pReporter,
  void* pUserData,
  int chngType
){
  XmlHash* pHash = (XmlHash*) pUserData;
  return pHash->trace(pCursor, pReporter, chngType);
}

/*
** This routine gets called whenever a trace is destroyed.  The
** user data for the trace being destroyed is passed in as the argument.
** The role of this routine is to delete the associated hash.
*/
static void hashTraceDeleteProc(void* pArg){
  XmlHash* pHash = (XmlHash*) pArg;
  char zHashName[500];
  snprintf(zHashName, sizeof zHashName, "-hash-%p-", pHash);

  /* This will take care of calling the destructor.
  */
  Tcl_DeleteCommand(pHash->pInterp, zHashName);
}

/*
** Called when the user asks for a new hash.  We create the xml trace
** ourself.
*/
PUBLIC XmlHash::XmlHash(
  Tcl_Interp *_pInterp, 
  const char* _zRoot, 
  int _iType,
  const char* _zName,
  short _bUnique,
  HashProcCtx* _pHashCtx
){
  pInterp = _pInterp;
  zRoot = new char[strlen(_zRoot)+1];
  strcpy(zRoot,_zRoot);
  iType = _iType;
  zName = new char[strlen(_zName)+1];
  strcpy(zName,_zName);
  pPrev = 0;
  pHashCtx = _pHashCtx;
  bUnique = _bUnique;

  /* Link ourself into the list.
  */
  pNext = pFirst;
  if( pFirst ) pFirst->pPrev = this;
  pFirst = this;

  /* Initialize the trace and our internal arrays.
  */
  XmlCursor cur(zRoot);
  assert(cur.ok());
  cur.trace(hashTraceCallback, this, XML_TRACE_ALL, hashTraceDeleteProc);

  visit(cur);
}

/*
** This is called whenever the trace is deleted or the user explicitly
** asks us to delete a hash.
*/
PUBLIC XmlHash::~XmlHash(){
  if( this==pFirst ){
    pFirst = pNext;
    if( pNext ) pNext->pPrev = 0;
  } else {
    if( pPrev ) pPrev->pNext = pNext;
    if( pNext ) pNext->pPrev = pPrev;
  }

  XmlCursor cur(zRoot);
  cur.untrace(hashTraceCallback, this, XML_TRACE_ALL);
  delete [] zRoot; zRoot = 0;
  delete [] zName; zName = 0;

  if( pHashCtx ){
    pHashCtx->invalidate();
    pHashCtx = 0;
  }
}

#if XMLHASH_STRICT_CHECKS && (!defined(NDEBUG) || !NDEBUG)
/*
** Re-read our state from the xml tree.  Abort if we're different than
** the xml tree thinks we should be.
*/
PRIVATE void XmlHash::StrictCheck(int chngType){

  /* If this method was called as a result of a DELETE trace, then our
  ** check will certainly fail, since the xml element is not actually deleted
  ** until after the traces have fired.
  */
  if( chngType==XML_TRACE_DELETE ){ return; }

  map_type tmp_map;
  path2value_type tmp_reverse_map;

  tmp_map = m_map;
  m_map.clear();
  tmp_reverse_map = m_reverse_map;
  m_reverse_map.clear();

  XmlCursor cur(zRoot);
  assert( cur.ok() );
  visit(cur);

  map_type::const_iterator p;
  path_set_type::const_iterator q;

  /* Check m_map.
  */
  for( p=m_map.begin(); p!=m_map.end(); ++p ){
    const char* zKey = p->first.c_str();
    const path_set_type& sPaths = p->second;
    
    map_type::const_iterator tmpIt = tmp_map.find(zKey);
    assert( tmpIt!=tmp_map.end() );
    const path_set_type& sTmpPaths = tmpIt->second;
    for( q=sPaths.begin(); q!=sPaths.end(); ++q ){
      assert( sTmpPaths.find(*q)!=sTmpPaths.end() );
    }
    for( q=sTmpPaths.begin(); q!=sTmpPaths.end(); ++q ){
      assert( sPaths.find(*q)!=sPaths.end() );
    }
  }
  for( p=tmp_map.begin(); p!=tmp_map.end(); ++p ){
    const char* zKey = p->first.c_str();
    const path_set_type& sPaths = p->second;
    
    map_type::const_iterator tmpIt = m_map.find(zKey);
    assert( tmpIt!=m_map.end() );
    const path_set_type& sTmpPaths = tmpIt->second;
    for( q=sPaths.begin(); q!=sPaths.end(); ++q ){
      assert( sTmpPaths.find(*q)!=sTmpPaths.end() );
    }
    for( q=sTmpPaths.begin(); q!=sTmpPaths.end(); ++q ){
      assert( sPaths.find(*q)!=sPaths.end() );
    }
  }

  /* Check m_reverse_map.
  */
  path2value_type::const_iterator revIt;
  for( revIt=m_reverse_map.begin(); revIt!=m_reverse_map.end(); ++revIt ){
    const char* zKey = revIt->first.c_str();
    const char* zVal = revIt->second.c_str();
    
    path2value_type::const_iterator tmpIt = tmp_reverse_map.find(zKey);
    assert( tmpIt!=tmp_reverse_map.end() );
    assert( strcmp(zVal, tmpIt->second.c_str())==0 );
  }
  for( revIt=tmp_reverse_map.begin(); revIt!=tmp_reverse_map.end(); ++revIt ){
    const char* zKey = revIt->first.c_str();
    const char* zVal = revIt->second.c_str();
    
    path2value_type::const_iterator tmpIt = m_reverse_map.find(zKey);
    assert( tmpIt!=m_reverse_map.end() );
    assert( strcmp(zVal, tmpIt->second.c_str())==0 );
  }
}
#endif /* XMLHASH_STRICT_CHECKS */

/*
** Write a description of ourself into pInterp.
*/
PUBLIC int XmlHash::describe(Tcl_Interp* pInterp){
  Tcl_ResetResult(pInterp);
  Tcl_AppendResult(pInterp, zRoot, " ", 
      iType==XMLHASH_TYPE_ATTR ? "attr" : "leaf", " ",
      zName, bUnique ? " -unique" : NULL, NULL);
  return TCL_OK;
}

/*
** Recursively visit all children of the node pointed to by cur, 
** initializing our arrays as we go.
*/
PRIVATE void XmlHash::visit(XmlCursor cur){
  addMap(GetSalientValue(&cur), cur.fullname());
  for( cur.move_first(); cur.ok(); cur.move_next() ){
    visit(cur);
  }
}

/*
** This is called whenever some interesting change happens to the xml
** tree.
*/
PUBLIC int XmlHash::trace(
  XmlCursor *pCursor,
  XmlCursor *pReporter,
  int chngType
){
  const char* zFullname = pCursor->fullname();
  const char* zCurVal = GetSalientValue(pCursor);
  const char* zOldVal = "";
  path2value_type::const_iterator i = m_reverse_map.find(zFullname);
  if( i!=m_reverse_map.end() ){
    zOldVal = i->second.c_str();
    removeMap(zOldVal, zFullname);
  }

  if( chngType!=XML_TRACE_DELETE ){ addMap(zCurVal, zFullname); }

  SANITY_CHECK(chngType);
  return 0;
}

/*
** Given a pointer to an xml node, return the salient value in the node.
** This is either the value of the attribute with name .zName, if we're
** an attribute hash, the value of the node itself if the node's name is
** .zName, or NULL otherwise.
*/
PRIVATE const char* XmlHash::GetSalientValue(const XmlCursor* pCur) const {
  switch( iType ){
    case XMLHASH_TYPE_ATTR:
      return pCur->getAttribute(zName);
    break;
    case XMLHASH_TYPE_LEAF:
      if( strcmp(pCur->name(), zName)==0 ){
        return pCur->get();
      }
    break;
  }
  return NULL;
}

/*
** Make zVal not point to zPath in m_map, and do the reverse
** in m_reverse_map.  If zVal is NULL, do nothing.
*/
PRIVATE void XmlHash::removeMap(const char* zVal, const char* zPath){
  if( zVal==NULL ){ return; }
  m_reverse_map.erase(zPath);
  m_map[zVal].erase(zPath);
  if( m_map[zVal].size()==0 ){ m_map.erase(zVal); }
}

/*
** Make zVal point to zPath in m_map, and do the reverse in 
** m_reverse_map.  If zVal is NULL, do nothing.
*/
PRIVATE void XmlHash::addMap(const char* zVal, const char* zPath){
  if( zVal==NULL ){ return; }
  assert( zPath && zPath[0] );
  m_map[zVal].insert(zPath);
  m_reverse_map[zPath] = zVal;
}

/*
** Dump our state to stderr.  Used for debugging.
*/
PUBLIC void XmlHash::dump() const {
  Tcl_Interp* pMyInterp = Tcl_CreateInterp();
  dump(pMyInterp);
  fprintf(stderr, "%s", Tcl_GetStringResult(pMyInterp));
  Tcl_DeleteInterp(pMyInterp);
}

/*
** Dump our state to the result field of the given interpreter.  Used
** for debugging.
*/
PUBLIC int XmlHash::dump(Tcl_Interp* pInterp) const {
  Tcl_ResetResult(pInterp);
  Tcl_AppendResult(pInterp, "forward:\n", NULL);
  map_type::const_iterator pElem;
  path_set_type::const_iterator p;
  for( pElem = m_map.begin(); pElem!=m_map.end(); ++pElem ){
    const char* zVal = (const char*) pElem->first.c_str();
    const path_set_type& sPaths = pElem->second;
    Tcl_AppendResult(pInterp, "  ", zVal, " -> ", bUnique ? "" : "\n", NULL);
    for( p = sPaths.begin(); p!=sPaths.end(); ++p ){
      const char* zPath = (const char*) p->c_str();
      Tcl_AppendResult(pInterp, bUnique ? "" : "    ", zPath, "\n", NULL);
    }
  }
  Tcl_AppendResult(pInterp, "backward:\n", NULL);
  path2value_type::const_iterator i;
  for( i = m_reverse_map.begin(); i!=m_reverse_map.end(); ++i ){
    const char* zPath = (const char*) i->first.c_str();
    const char* zVal = (const char*) i->second.c_str();
    Tcl_AppendResult(pInterp, "  ", zPath, " -> ", zVal, "\n", NULL);
  }
  return TCL_OK;
}

/*
** Return a list of all hashed values with their corresponding elements.
**
** The list is of the form VAL1 LIST1 VAL2 LIST2...,
** where the elements in LISTi all have salient value VALi.
**
** If this is a unique-hash and any LISTi would have more than one element,
** we abort execution of the application.
*/
PUBLIC int XmlHash::list(Tcl_Interp* pInterp) const {
  Tcl_ResetResult(pInterp);
  Tcl_Obj* pList = Tcl_NewListObj(0, NULL);
  map_type::const_iterator i;
  for( i=m_map.begin();i!=m_map.end();++i ){
    const char* zVal = (const char*) i->first.c_str();
    Tcl_ListObjAppendElement(pInterp, pList, Tcl_NewStringObj(zVal, -1));
    Tcl_ListObjAppendElement(pInterp, pList, find(zVal));
  }
  Tcl_SetObjResult(pInterp, pList);
  return TCL_OK;
}

/*
** Dump some debugging information and, if this is a debugging build,
** abort execution of the application.
**
** This is called when we try to get multiple elements out of a unique-hash.
**
** This method never returns unless NDEBUG is defined.
*/
PRIVATE void XmlHash::fail(
  const char* zVal,
  const char* zElt1,
  const char* zElt2 
) const {
  XmlCursor cur(zRoot);
  fprintf(stderr, "Tried to get at least two elements out of unique-hash\n"
      "  tree: %s,\n"
      "  root: %s,\n"
      "  type: %s,\n"
      "  name: %s,\n"
      "  value: %s,\n"
      "  elements: %s %s\n", 
      cur.render(0), zRoot, 
      (iType==XMLHASH_TYPE_LEAF ? "leaf" : "attr"), zName, zVal, zElt1, zElt2);
  Tcl_Eval(pInterp, "::debug::probe");
  assert(0 && "Tried to get multiple elements out of a unique-hash.");
}

/*
** Return a list of the full paths of the xml elements whose salient 
** values given this hash are zVal.
**
** If this is a unique hash, don't return a list, just return "" or the
** appropriate element.
**
** If we're a unique hash while multiple nodes have salient value zVal,
** abort if this is a debug build and just pick one to return otherwise.
*/
PUBLIC Tcl_Obj* XmlHash::find(const char* zVal) const {
  map_type::const_iterator i = m_map.find(zVal);
  if (i==m_map.end()){ return Tcl_NewStringObj("", -1); }
  const path_set_type& sPaths = i->second;
  assert( sPaths.size()>0 );
  if( bUnique ){
    if( sPaths.size()>1 ){
      fail(zVal, sPaths.begin()->c_str(), (++sPaths.begin())->c_str());
    }
    return Tcl_NewStringObj(sPaths.begin()->c_str(), -1);
  } else {
    Tcl_Obj* pList = Tcl_NewListObj(0, NULL);
    Tcl_Obj* pListElem;
    path_set_type::const_iterator it;
    for( it = sPaths.begin(); it!=sPaths.end(); ++it ){
      const char* zPath = it->c_str();
      pListElem = Tcl_NewStringObj(zPath, -1);
      Tcl_ListObjAppendElement(pInterp, pList, pListElem);
    }
    return pList;
  }
}

/*
** Leave the given cursor pCur pointing to the xml element whose salient 
** value given this hash is zVal.
**
** If more than one such element exists, just leave pCur pointing to
** one of them.  Except, if this is a unique hash, we abort execution
** of the application.
**
** If no such element exists, pCur is left unchanged.
**
** RETURN VALUE: 1 if an element with salient value zVal exists and
** pCur is non-null; 0 otherwise.
*/
PUBLIC int XmlHash::find(const char* zVal, XmlCursor* pCur) const {
  if( pCur==NULL ){ return 0; }
  map_type::const_iterator i = m_map.find(zVal);
  if( i==m_map.end() ){ return 0; }
  const path_set_type& sPaths = i->second;
  if( sPaths.size()==0 ){ return 0; }
  if( bUnique && sPaths.size()>1 ){
    fail(zVal, sPaths.begin()->c_str(), (++sPaths.begin())->c_str());
  }
  if( !pCur->move(sPaths.begin()->c_str()) ){
    assert(0 && "can't happen");
  }
  return 1;
}

/*
** Change the value of a dynamically allocated string.  Return TRUE if
** there is a change and return FALSE if the old and new values were
** the same.
*/
static int ChangeStr(char **pzStr, const char *zNew){
  if( zNew==0 && *pzStr==0 ) return 0;
  if( zNew && *pzStr && strcmp(zNew, *pzStr)==0 ) return 0;
  if( *pzStr ){
    ckfree(*pzStr);
    *pzStr = 0;
  } 
  if( zNew ){
    *pzStr = (char*) ckalloc(strlen(zNew)+1);
    if( *pzStr ) strcpy(*pzStr, zNew);
  }
  return 1;
}

/* -------------------------- HASH PROCS ------------------------------ */

#if INTERFACE
/*
** class: HashProcCtx
**
** Maintains some stuff a hash procedure needs.  A hash procedure is what
** hash.tcl used to implement; i.e., it monitors the existence of an 
** underlying XmlHash, which can be transient.  If a caller wants data out 
** of a hash with particular properties, we instantiate one if necessary.
**
** The tcl implementation, while it ran in about 200 usec/iteration, was
** too slow (since it's called a lot).  The C implementation runs in about
** 30 usec/iteration, under profiling.
**
** See common/misc/hash.tcl for more details on how the interface works.
*/
class HashProcCtx {
public:
  Tcl_Interp* pInterp;
  char* zRoot;      // XML root of the hash.
  int iType;        // XMLHASH_TYPE_LEAF or XMLHASH_TYPE_ATTR.
  char* zName;      // Name of the Tcl hash-command.
  int bUnique;      // TRUE if this is a unique hash
  char* zProcName;  // Name of the "external" Tcl command.
                    // Used for debugging purposes only

  XmlHash* pHash;
  Tcl_ObjCmdProc* pObjProc;
  char* zCmdName;

};
#endif /* INTERFACE */

/*
** Constructor/destructor.
*/
PUBLIC HashProcCtx::HashProcCtx(
  Tcl_Interp* _pInterp,
  const char* _zRoot,
  int _iType,
  const char* _zName,
  int _bUnique,
  const char* _zProcName
){
  assert(_pInterp!=0);
  pInterp = _pInterp;
  zProcName = zRoot = zName = zCmdName = 0;
  pHash = 0;
  pObjProc = 0;
  ChangeStr(&zRoot, _zRoot);
  ChangeStr(&zName, _zName);
  ChangeStr(&zProcName, _zProcName);
  assert(_iType==XMLHASH_TYPE_LEAF || _iType==XMLHASH_TYPE_ATTR);
  iType = _iType;
  bUnique = _bUnique;
}
PUBLIC HashProcCtx::~HashProcCtx(){
  if( zCmdName ){ Tcl_DeleteCommand(pInterp, zCmdName); }
  ChangeStr(&zRoot, NULL);
  ChangeStr(&zName, NULL);
  ChangeStr(&zCmdName, NULL);
  ChangeStr(&zProcName, NULL);
}

/* Some forward declarations.
*/
static int HashProcCmdProc(ClientData clientdata, Tcl_Interp *pInterp,
  int objc, Tcl_Obj *const* objv);
static void HashProcDeleteProc(ClientData clientdata);

/* 
** Called by the pHash object, when it's being deleted.
*/
PUBLIC void HashProcCtx::invalidate() {pHash = 0;}

/*
** tclcmd: create_hash_proc PNAME ROOT TYPE NAME ?-unique?
** title:  Create a hash accessor tcl command.
*/
static int XmlCreateHashProcCmd(
  ClientData NotUsed,      // Not used 
  Tcl_Interp *pInterp,     // The interpreter controlling the new hash proc
  int objc,                // Number of arguments
  Tcl_Obj *const* objv     // Value of each argument
){
  const char* zProcName;
  const char* zRoot;
  const char* zType;
  const char* zName;
  const char* zUnique;

  int iType;
  int bUnique = 0;
  if( objc!=5 && objc!=6 ){
    Tcl_WrongNumArgs(pInterp, 1, objv, "PNAME ROOT TYPE NAME ?-unique?");
    return TCL_ERROR;
  }
  zProcName = Tcl_GetStringFromObj(objv[1], 0);
  zRoot = Tcl_GetStringFromObj(objv[2], 0);
  zType = Tcl_GetStringFromObj(objv[3], 0);
  if( zType[0]=='l' && strcmp(zType, "leaf")==0 ){
    iType = XMLHASH_TYPE_LEAF;
  } else if( zType[0]=='a' && strcmp(zType, "attr")==0 ){
    iType = XMLHASH_TYPE_ATTR;
  } else {
    Tcl_ResetResult(pInterp);
    Tcl_AppendResult(pInterp, "Invalid hash type ", zType, 
        ": must be leaf or attr", NULL);
    return TCL_ERROR;
  }
  zName = Tcl_GetStringFromObj(objv[4], 0);
  if( objc==6 ){
    zUnique = Tcl_GetStringFromObj(objv[5], 0);
    if( zUnique[0]=='-' && strcmp(zUnique, "-unique")==0 ){
      bUnique = 1;
    } else if( !zUnique[0] ){
      bUnique = 0;
    } else {
      Tcl_ResetResult(pInterp);
      Tcl_AppendResult(pInterp,
          "Invalid argument ", zUnique, ": must be -unique or {}", NULL);
      return TCL_ERROR;
    }
  }

  HashProcCtx *pCtx = new HashProcCtx(pInterp, zRoot, iType, zName, bUnique,
      zProcName);
  Tcl_CreateObjCommand(pInterp, zProcName, HashProcCmdProc, pCtx,
                       HashProcDeleteProc);

  return TCL_OK;
}

static int CreateHash(
  Tcl_Interp* interp,
  const char* zRoot,
  int iType,
  const char* zName,
  bool bUnique,
  HashProcCtx* pHashCtx);

/*
** This proc is called when a caller actually says, e.g., 
** "::hash::softools arg arg...".  Instantiate an underlying XmlHash
** if necessary and pass the args downward.
*/
static int HashProcCmdProc(
  ClientData clientdata,   // Pointer to a HashProcCtx instance.
  Tcl_Interp *pInterp,     // The interpreter controlling the new cursor
  int objc,                // Number of arguments
  Tcl_Obj *const* objv     // Value of each argument
){
  HashProcCtx* pCtx = (HashProcCtx*) clientdata;
  if( pCtx->pHash==NULL ){
    int rc = CreateHash(pInterp,
        pCtx->zRoot, pCtx->iType, pCtx->zName, pCtx->bUnique, pCtx);
    if( rc==-1 ){

      /* pCtx->zRoot doesn't exist at the moment, but we don't
      ** consider that an error.
      */
      Tcl_ResetResult(pInterp);
      return TCL_OK;
    } else if( rc!=TCL_OK ){
      return rc;
    }

    const char* zHashName = Tcl_GetStringResult(pInterp);
    Tcl_CmdInfo info;
    if( Tcl_GetCommandInfo(pInterp, zHashName, &info)==0 ){
      assert(0 && "can't happen");
      Tcl_ResetResult(pInterp);
      return TCL_OK;
    }
    ChangeStr(&pCtx->zCmdName, zHashName);
    pCtx->pHash = (XmlHash*) info.objClientData;
    pCtx->pObjProc = info.objProc;
  }

  /* Defer to the underlying hash's implementation.
  ** Calls HashCmdProc().
  */
  return (*pCtx->pObjProc)(pCtx->pHash, pInterp, objc, objv);
}

/*
** Called when a hash proc is about to be destroyed.  Delete the associated
** context.
*/
static void HashProcDeleteProc(
  ClientData clientdata   // Pointer to a HashProcCtx instance.
){
  HashProcCtx* pCtx = (HashProcCtx*) clientdata;
  delete pCtx; pCtx = 0;
}

/* ----------------------- 'xml hash' SUBCOMMAND ------------------------- */

/* Forward declarations.
*/
static int HashCmdProc(ClientData, Tcl_Interp*, int, Tcl_Obj *const*);
static void HashDeleteProc(ClientData);

/*
** This routine processes the "hash" method of an xml cursor. This
** routine is called by XmlCmdProc when the first argument is "hash".
*/
int HashProc(
  Tcl_Interp *interp,     //Interpreter controlling the cursor
  int objc,               //Number of arguments
  Tcl_Obj *const* objv,   //Value of each argument
  XmlCursor *pCur         //Pointer to the cursor object
){
  static const char *ATTR_strs[] = {
    "create",              "destroy",               "list",   0
  };
  enum ATTR_enum {
    HASH_CREATE,           HASH_DESTROY,            HASH_LIST,
  };
  int index;

  if( Tcl_GetIndexFromObj(interp, objv[2], ATTR_strs, "option", 0, &index)){
    return TCL_ERROR;
  }
  switch( (enum ATTR_enum)index ){
    /*
    ** xmlp-tcl-cmd: CURSOR hash create ROOT TYPE NAME ?-unique?
    **
    ** Create a hash (which is an index mapping values V to a list of xml 
    ** nodes with salient value = V).  See above for the definition of 
    ** "salient value"; it depends on how the new hash is defined.
    **
    ** Initialize the hash and return a handle to the hash.  The handle 
    ** is the hex representation of the memory location of the XmlHash 
    ** instance implementing the hash.
    */
    case HASH_CREATE: {
      short bUnique = 0;
      if( objc<6 || objc>7 ){
        Tcl_WrongNumArgs(interp, 3, objv, "ROOT TYPE NAME ?-unique?");
        return TCL_ERROR;
      }
      if( objc==7 ){
        if( strcmp(Tcl_GetStringFromObj(objv[6], 0), "-unique")!=0 ){
          Tcl_ResetResult(interp);
          Tcl_AppendResult(interp, "wrong args: must be CURSOR hash create"
              " ROOT TYPE NAME ?-unique?", 0);
          return TCL_ERROR;
        }
        bUnique = 1;
      }
      const char* zRoot = Tcl_GetStringFromObj(objv[3], 0);
      const char* zType = Tcl_GetStringFromObj(objv[4], 0);
      int iType;
      const char* zName = Tcl_GetStringFromObj(objv[5], 0);
      if( strcmp(zType, "attr")==0 ){
        iType = XMLHASH_TYPE_ATTR;
      } else if( strcmp(zType, "leaf")==0 ){
        iType = XMLHASH_TYPE_LEAF;
      } else {
        Tcl_ResetResult(interp);
        Tcl_AppendResult(interp, "invalid type \"", zType, "\": must be ",
            "\"attr\" or \"leaf\"", NULL);
        return TCL_ERROR;
      }
      int rc = CreateHash(interp, zRoot, iType, zName, bUnique, 0);
      return (rc==-1) ? TCL_ERROR : rc;
      break;
    }
    
    /*
    ** xmlp-tcl-cmd: CURSOR hash destroy HANDLE
    */
    case HASH_DESTROY: {
      if( objc!=4 ){
        Tcl_WrongNumArgs(interp, 3, objv, "HANDLE");
        return TCL_ERROR;
      }
      const char* zHashName = Tcl_GetStringFromObj(objv[3], NULL);
      Tcl_DeleteCommand(interp, zHashName);
      Tcl_ResetResult(interp);
      return TCL_OK;
      break;
    }

    /*
    ** xmlp-tcl-cmd: CURSOR hash list
    **
    ** Return a list of extant hashes.
    */
    case HASH_LIST: {
      if( objc!=3 ){
        Tcl_WrongNumArgs(interp, 3, objv, "");
        return TCL_ERROR;
      }
      return XmlHash::dump_list(interp);
    }
  }
  return TCL_OK;
}

/*
** This routine runs when the Tcl interpreter executes one of the
** commands associated with an XML hash.  This routine figures out
** which method was invoked and takes appropriate action.
*/
static int HashCmdProc(
  ClientData pData,       // A pointer to an XmlHash
  Tcl_Interp *interp,     // The interpreter controlling the new cursor
  int objc,               // Number of arguments
  Tcl_Obj *const* objv    // Value of each argument
){
  static const char *HASH_strs[] = {
    "all",                      "describe",                 "destroy",
    "dump",                     "get",
    0
  };
  enum HASH_enum {
    HASH_ALL,                   HASH_DESCRIBE,              HASH_DESTROY,
    HASH_DUMP,                  HASH_GET,
  };
  XmlHash *pHash = (XmlHash*)pData;
  if( !pHash ){
    assert(0 && "can't happen");
    Tcl_SetResult(interp, (char *) "Invalid hash", TCL_STATIC);
    return TCL_ERROR;
  }
  int index;

  if( objc<2 ){
    Tcl_WrongNumArgs(interp, 1, objv, "METHOD ?ARG ...?");
    return TCL_ERROR;
  }
  if( Tcl_GetIndexFromObj(interp, objv[1], HASH_strs, "option", 0, &index)){
    return TCL_ERROR;
  }
  switch( (enum HASH_enum)index ){
    case HASH_ALL: {
      if( objc!=2 ){
        Tcl_WrongNumArgs(interp, 2, objv, "");
        return TCL_ERROR;
      }
      return pHash->list(interp);
      break;
    }
    case HASH_DESCRIBE: {
      if( objc!=2 ){
        Tcl_WrongNumArgs(interp, 2, objv, "");
        return TCL_ERROR;
      }
      return pHash->describe(interp);
      break;
    }
    case HASH_DESTROY: {
      if( objc!=2 ){
        Tcl_WrongNumArgs(interp, 2, objv, "");
        return TCL_ERROR;
      }
      char zNameBuf[100];
      snprintf(zNameBuf, sizeof zNameBuf, "-hash-%p-", pHash);
      Tcl_DeleteCommand(interp, zNameBuf);
      Tcl_ResetResult(interp);
      return TCL_OK;
      break;
    }
    case HASH_DUMP: {
      if( objc!=2 ){
        Tcl_WrongNumArgs(interp, 2, objv, "");
        return TCL_ERROR;
      }
      pHash->dump();
      return pHash->dump(interp);
      break;
    }
    case HASH_GET: {
      if( objc!=3 ){
        Tcl_WrongNumArgs(interp, 2, objv, "KEY");
        return TCL_ERROR;
      }
      const char* zKey = Tcl_GetStringFromObj(objv[2], 0);
      Tcl_SetObjResult(interp, pHash->find(zKey));
      return TCL_OK;
      break;
    }
  }
  return TCL_OK;
}

/*
** This function is called when the tcl procedure associated with a hash
** is destroyed.  We destroy the hash itself.
*/
static void HashDeleteProc(ClientData pData){
  XmlHash* pHash = (XmlHash*) pData;
  delete pHash; pHash = 0;
}

/*
** Create a new hash object and associate it with an automatically-generated
** Tcl command name.
**
** If zRoot is not currently a valid xml path, return -1.  For any other
** error, return TCL_ERROR.  On success, return TCL_OK.
*/
static int CreateHash(
  Tcl_Interp* interp,
  const char* zRoot,
  int iType,
  const char* zName,
  bool bUnique,
  HashProcCtx* pHashCtx
){
  XmlCursor cur(zRoot);
  if( !cur.ok() ){
    Tcl_AppendResult(interp, zRoot, ": no such xml element", 0);
    return -1;
  }
  XmlHash* pHash = new XmlHash(interp, zRoot, iType, zName, bUnique,
      pHashCtx);
  if( pHash==0 ){
    Tcl_SetResult(interp, (char *) "unable to create the new hash", 
        TCL_STATIC);
    return TCL_ERROR;
  }
  char zNameBuf[100];
  snprintf(zNameBuf, sizeof zNameBuf, "-hash-%p-", pHash);
  Tcl_CreateObjCommand(interp, (char*)zNameBuf, HashCmdProc, (void*)pHash,
                       HashDeleteProc);
  Tcl_SetResult(interp, zNameBuf, TCL_VOLATILE);
  return TCL_OK;
}

/* ------------------------ COMMAND REGISTRATION -------------------------- */

/*
** Register some commands with the tcl interpreter.
*/
extern "C" int Xmlp_Hash_Init(Tcl_Interp* interp){
  Tcl_CreateObjCommand(interp, "create_hash_proc", XmlCreateHashProcCmd, 0, 0);
  return TCL_OK;
}
